{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    Copyright (C) 2006

  ---------------------------------------------------------------------------

    20.07.2006 - V0.20
      First public release. For testing and debuging...
    
    03.09.2006 - V0.21
      Improved date/time support for lazarus, some bugs fixed with queries,
      fielddefs...
      Added partial support  for Kexi system tables in TSqlitePassTranslator_Kexi
      and more...

    24.09.2006 - V0.22
      TSqlitePassTranslator_Kexi.GetDatabaseObjects
       The SQL statement is correct now.
      TSqlitePassDataset.InternalInitFieldDefs
       It uses now the new TSqlitePassRecordsetCache.HasPrimaryKey property
       to return the correct fieldDefs.
      TSqliteIndexDefs - TSqliteIndex - TSqliteIndexColumns - TSqliteIndexColumn
       Partially implemented
       Added TSqlitePassIndexColumn.Position property

    17/10/2006 - V0.23
      TSqlitePassTranslator.GetIndexDefs;
       IndexDefs[i].Sql corrected to FDatabase.IndexDefs[i].Sql
      Added new procedure CreateIndex(Index: TSqlitePassDatasetIndex); overload;
      and many bug fixes.
      Added basic TSqlitePassViews implementation and Database.Views property
      Changes made in TSqliteDataset.internalOpen to handle views.

    18/12/2006 - V0.24
      Added TSqlitePassDatabasesAttached to maintain a list of external databases
      attached to the current database.
               -> Attached Tables support
               -> Attached Queries support
      Rework on date/time support for lazarus since TDataset.GetFieldData
      implementation has changed with FPC 2.0.4 / Lazarus 0.9.18

  --------------------------------------------------------------------------- }


unit SqlitePassDbo;
{$i SqlitePassDbo.inc}

Interface
uses
{$IFNDEF FPC}
DBCommon,
{$ENDIF}
Classes, Windows, SysUtils, Db, Forms, Math,
StdCtrls, // to use a TMemo as dirty debug screen
SqlitePassUtils, SqlitePassApi_v3, SqlitePassErrorLang, SqlitePassEngine;

Const
{ TSqliteDataset }
SqlitePasBookMarkSize=SizeOf(Integer);

{ Arbitrary index for temporaly insert/append operation }
NewTempRecordIndex = -9999;



Type
TSqlitePassDataset = class;

 PPointer= ^Pointer;
 TSqlitePassExecCallBack=function(acolumns: Integer; aColumnValues, aColumnNames:Ppchar):integer of object;
 PSqlitePassExecCallBack=^TSqlitePassExecCallBack;
 TSqlitePassGetPragmaCallback=function(UserData: Pointer; ColumnCount: Integer; ColumnValues, ColumnNames: PPointer): Integer; cdecl;
 TSqlitePassDatabaseObjectTypes =(dboAll, dboTableDefs, dboQueryDefs, dboIndexDefs, dboViews, dboTriggers);
 TSqlitePassDatasetType = (dtTable, dtQuery, dtView, dtUnknown);


{Record to temporary store information about FielDefs when we extract
 them from Sqlite engine
 ----------------------------------------------------------------------- }
PSqliteFieldDef=^TSqliteFieldDef;
TSqliteFieldDef=record
  ColIdx, Name, FieldType, Required, DefaultValue, PrimaryKey: PChar;
  end;

{ Information about one of the records filled in the dataset records buffers
  TBufferList = array[0..dsMaxBufferCount - 1] of PChar; (defined in the db unit)
  the TBookmarkFlag = (bfCurrent, bfBOF, bfEOF, bfInserted) is also defined in
  the db unit. FPC Uses this Flag via the GetBookmarkFlag and SetBookmarkFlag
  methods in order to know 'where' is located a specific record.
 ----------------------------------------------------------------------- }

 TLoadMode = (lmAppend, lmEmptyAppend);

 TSqlitePassDatabase=Class;
 TSqlitePassDatabaseAttached=Class;

 TSqlitePassFieldDefParser = Class
 Public
 Name: String;
 FieldType: TFieldType;
 end;

{ TSqliteAttachedDatabases : Maintain a list of all attached databases }

TSqlitePassDatabasesAttached = Class(TSqlitePassObjectList)
  Private
   FDatabase: TSqlitePassDatabase;
  Protected
    Function GetItem(Index: Integer): TSqlitePassDatabaseAttached;
    Procedure SetItem(Index: Integer; Database: TSqlitePassDatabaseAttached);
  public
    destructor Destroy; override;
    constructor Create(Owner: TSqlitePassDatabase);
    function DatabaseByName(Const Name: String): TSqlitePassDatabaseAttached;
    function FindDatabase(const Value: string): TSqlitePassDatabaseAttached;
    property Items[Index: Integer]: TSqlitePassDatabaseAttached read GetItem write SetItem; default;
    property Count;
  end;

{ TableDefs }

TSqlitePassTableDefs = class;

TSqlitePassTableDef = class
  private
  FTableName: String;
  FTableNameExt: String;
  FSql: String;
  FFieldDefs: TFieldDefs;
  FTableDefs: TSqlitePassTableDefs;
  public
  Constructor Create(Owner: TSqlitePassTableDefs);
  Destructor Destroy; override;
  procedure DeleteTable(Silent: Boolean = False);
  procedure RenameAs(NewName: String);
  procedure Reindex;
  property TableName: String Read FTableName;
  property TableNameExt: String Read FTableNameExt;
  property Sql: String Read FSql;
  property Fields: TFieldDefs Read FFieldDefs;
  end;

TSqlitePassTableDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassTableDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassTableDef);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  destructor Destroy;
  procedure CreateTable(Sql: String);
  procedure DeleteTable(TableName: String; Silent: Boolean = False);
  function TableByName(const Value: string): TSqlitePassTableDef;
  function FindTable(const Value: string): TSqlitePassTableDef;
  property Attached: Boolean Read GetAttached;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassTableDef read GetItem write SetItem; default;
  end;


  { QueryDefs }

TSqlitePassQueryDefs = class;

TSqlitePassQueryDef = class
  private
  FQueryName: String;
  FQueryNameExt: String;
  FSql: String;
  FFieldDefs: TFieldDefs;
  FQueryDefs: TSqlitePassQueryDefs;
  FParams: Tparams;
  function GetAttached: Boolean;
  public
  Constructor Create(Owner: TSqlitePassQueryDefs);
  Destructor Destroy; override;
  property Attached: Boolean Read GetAttached;
  procedure DeleteQuery;
  procedure RenameAs(NewName: String);
  property QueryName: String Read FQueryName;
  property QueryNameExt: String Read FQueryNameExt;
  property Sql: String Read FSql;
  property Params: TParams Read FParams;
  property Fields: TFieldDefs Read FFieldDefs;
  end;

TSqlitePassQueryDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassQueryDef;
  procedure SetItem(Index: Integer; const Value: TSqlitePassQueryDef);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  procedure CreateQuery(Sql: String);
  property Attached: Boolean Read GetAttached;
  procedure DeleteQuery(QueryName: String);
  function QueryByName(const Value: string): TSqlitePassQueryDef;
  function FindQuery(const Value: string): TSqlitePassQueryDef;
  property Count;
  procedure Refresh;
  property Items[Index: Integer]: TSqlitePassQueryDef read GetItem write SetItem; default;
  end;



{ Indexes management }
TSqlitePassIndexColumns = class;
TSqlitePassIndex = class;
TSqlitePassDatabaseIndexDefs = class;
TSqlitePassDatasetIndexDefs = class;

TSqlitePassIndexDirection = (cidAscending, cidDescending, cidUnknown);

TSqlitePassIndexColumn = class
 public
 Position: Integer;
 CollatingSequence: String;
 ColumnName: String;
 Direction: TSqlitePassIndexDirection;
 end;

TSqlitePassIndexColumns = class(TSqlitePassObjectList)
  private
  FIndex: TSqlitePassIndex;
  function GetItem(Index: Integer): TSqlitePassIndexColumn;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndexColumn);
  public
  constructor Create(Owner: TSqlitePassIndex);
  function ColumnByName(const Value: string): TSqlitePassIndexColumn;
  function FindColumn(const Value: string): TSqlitePassIndexColumn;
  property Items[Index: Integer]: TSqlitePassIndexColumn read GetItem write SetItem; default;
  property Count;
  end;

TSqlitePassIndex = class
 private
  FDatasetName: String;
  FIndexName: String;
  FOverwriteIfExists: Boolean;
  FColumns : TSqlitePassIndexColumns;
  FSql: String;
  FUnique : Boolean;
 public
  constructor Create;
  Destructor Destroy; override;
  property Sql: String Read FSql;
  property TableName: String Read FDatasetName Write FDatasetName;
  property IndexName: String Read FIndexName Write FIndexName;
  property Columns: TSqlitePassIndexColumns read FColumns write FColumns;
  property Unique: Boolean read FUnique write FUnique;
 end;

TSqlitePassDatabaseIndex = class(TSqlitePassIndex)
 private
  FIndexDefs: TSqlitePassDatabaseIndexDefs;
  function GetAttached: Boolean;
  public
  constructor Create(Owner: TSqlitePassDatabaseIndexDefs);
  procedure Reindex;
  procedure DeleteIndex;
  property Attached: Boolean Read GetAttached;
  end;

TSqlitePassDatasetIndex = class(TSqlitePassIndex)
 private
  FIndexDefs: TSqlitePassDatasetIndexDefs;
  function GetAttached: Boolean;
  public
  constructor Create(Owner: TSqlitePassDatasetIndexDefs);
  procedure Reindex;
  procedure DeleteIndex;
  property Attached: Boolean Read GetAttached;
  end;

TSqlitePassDatabaseIndexDefs = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassIndex;
  procedure SetItem(Index: Integer; const Value: TSqlitePassIndex);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  function IndexByName(const Value: string): TSqlitePassIndex;
  function FindIndex(const Value: string): TSqlitePassIndex;
  procedure CreateIndex(Sql: String); overload;
  procedure CreateIndex(Index: TSqlitePassDatasetIndex); overload;
  procedure DeleteIndex(IndexName: String);
  procedure Reindex;
  procedure Refresh;
  procedure RenameIndex(OldIndexName, NewIndexName: String);
  property Attached: Boolean Read GetAttached;
  property Items[Index: Integer]: TSqlitePassIndex read GetItem write SetItem; default;
  property Count;
  end;

TSqlitePassDatasetIndexDefs = class(TSqlitePassDatabaseIndexDefs)
  private
  FDataset: TSqlitePassDataset;
  public
  constructor Create(Owner: TSqlitePassDataset);
  procedure Refresh;
  procedure RefreshFromCache;
  end;

{ Views }

TSqlitePassViews = class;

TSqlitePassView = class
 private
 FViewName: String;
 FTemp: Boolean;
 FSqlCreateStmt: String;
 FSqlSelectStmt: String;
 FViews: TSqlitePassViews;
 function GetAttached: Boolean;
 public
 Constructor Create(Owner: TSqlitePassViews);
 Destructor Destroy; override;
 procedure DeleteView;
 property Attached: Boolean Read GetAttached;
 property ViewName: String Read FViewName;
 property SqlCreateStmt: String Read FSqlCreateStmt;
 property SqlSelectStmt: String Read FSqlSelectStmt;
 end;

TSqlitePassViews = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  function GetAttached: Boolean;
  function GetItem(Index: Integer): TSqlitePassView;
  procedure SetItem(Index: Integer; const Value: TSqlitePassView);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  procedure CreateView(Sql: String);
  procedure DeleteView(ViewName: String);
  function ViewByName(const Value: string): TSqlitePassView;
  function FindView(const Value: string): TSqlitePassView;
  property Count;
  procedure Refresh;
  property Attached: Boolean Read GetAttached;
  property Items[Index: Integer]: TSqlitePassView read GetItem write SetItem; default;
  end;


{ Triggers }

TSqlitePassTriggers = class;

TSqlitePassTrigger = class
 private
  FTemporary: Boolean;
  FTriggerName: String;
  FTriggers: TSqlitePassTriggers;
  FSql: String;
  FIndexName: String;
  function GetAttached: Boolean;
 public
  constructor Create(Owner: TSqlitePassTriggers);
  Destructor Destroy; override;
  procedure DeleteTrigger;
  property Attached: Boolean Read GetAttached;
  property Sql: String Read FSql;
  property Temporary: Boolean Read FTemporary Write FTemporary;
  property TriggerName: String Read FIndexName Write FIndexName;
 end;

TSqlitePassTriggers = class(TSqlitePassObjectList)
  private
  FDatabase: TSqlitePassDatabase;
  FAttachedDatabase: TSqlitePassDatabaseAttached;
  function GetAttached: Boolean;
  Function GetItem(Index: Integer): TSqlitePassTrigger;
  Procedure SetItem(Index: Integer; Value: TSqlitePassTrigger);
  public
  constructor Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
  procedure CreateTrigger(Sql: String);
  procedure DeleteTrigger(TriggerName: String);
  function TriggerByName(const Value: string): TSqlitePassTrigger;
  function FindTrigger(const Value: string): TSqlitePassTrigger;
  property Attached: Boolean Read GetAttached;
  property Items[Index: Integer]: TSqlitePassTrigger read GetItem write SetItem; default;
  property Count;
  procedure Refresh;
  end;


{TSqlitePassDatasets : Maintain a list of all ASSIGNED/ACTIVE datasets in the database}

 TSqlitePassDatasets = Class(TSqlitePassObjectList)
  Private
    FDatabase: TSqlitePassDatabase;
  Protected
    Function GetItem(Index: Integer): TSqlitePassDataset;
    Procedure SetItem(Index: Integer; Dataset: TSqlitePassDataset);
  public
    destructor Destroy; override;
    constructor Create(Owner: TSqlitePassDatabase);
    function DatasetByName(Const Name: String): TSqlitePassDataset;
    function FindDataset(const Value: string): TSqlitePassTrigger;
    property Items[Index: Integer]: TSqlitePassDataset read GetItem write SetItem; default;
    property Count;
  end;

TSqlitePassRecordsetCache=class;

TSqlitePassTranslator = class
  Private
    { Used to retrieve infos about TableDefs, indexDefs... }
    FTempDataset: TSqlitePassRecordsetCache;
    { Used to save and restore date/time format settings }
    FDatabase: TSqlitePassDatabase;
    FDateSeparator: Char;
    FShortDateFormat: String;
    FTimeSeparator: Char;
    FLongTimeFormat: String;
    FDecimalSeparator: Char;
    FOldDateSeparator: Char;
    FOldShortDateFormat: String;
    FOldTimeSeparator: Char;
    FOldLongTimeFormat: String;
    FOldDecimalSeparator: Char;
    procedure RestoreFloatFormatSettings;
    procedure SetFloatFormatSettings;
  protected
    //*** Date Time Utilities ***********************************************
    procedure RestoreDateFormatSettings;
    procedure RestoreTimeFormatSettings;
    procedure SetDateFormatSettings;
    procedure SetTimeFormatSettings;
    //*** GetFieldData ******************************************************
    function StrToBooleanFieldBuffer(Const StrValue: String; out Value: WordBool): Boolean;
    function StrToIntFieldBuffer(Const StrValue: String; out Value: Integer): Boolean;
    function StrToFloatFieldBuffer(Const StrValue: String; out Value: Double): Boolean;
    function StrToDateFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
    function StrToTimeFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
    function StrToDateTimeFieldBuffer(StrValue: String; out Value: TDateTimeRec): Boolean;
    function StrToFieldBuffer(Const StrValue: String; out Value: Pointer): Boolean;
    function StrToStrFieldBuffer(Const StrValue: String; out Value: String): Boolean;
    function StrToByteFieldBuffer(Const StrValue: String; out Value: String): Boolean;
    function StrToBlobFieldBuffer(Const Dataset: TSqlitePassDataset; Const StrValue: String; out Value: Pointer): Boolean;
    //*** SetFieldData ******************************************************
    procedure BooleanFieldBufferToStr(Const Value: Boolean; out StrValue: String);
    procedure IntFieldBufferToStr(Const Value: Integer; out StrValue: String);
    procedure FloatFieldBufferToStr(Const Value: Double; out StrValue: String);
    procedure DateFieldBufferToStr(Const Value: TDateTimeRec; out StrValue: String);
    procedure TimeFieldBufferToStr(Const Value: TDateTimeRec; out StrValue: String);
    procedure DateTimeFieldBufferToStr(Const Value: TDateTimeRec; out StrValue: String);
    procedure FieldBufferToStr(Const Value: PChar; out StrValue: String);
    procedure ByteFieldBufferToStr(Const Value: PChar; out StrValue: String);
    //*** Bind Values to Sql statement before Dataset.post
    procedure BindValues(const fst: pointer; const FieldDefs: TFieldDefs; const Rec: TMemRecord);
    function BindValue(const fst: Pointer; const FieldDef: TFieldDef; const FieldValue: String): Boolean;
  public
    Constructor Create(Owner: TSqlitePassDatabase);
    Destructor Destroy;override;
    function GetFieldData(Const Dataset: TSqlitePassDataset; Const StrValue: String; Const DataType: TFieldType; FieldBuffer: Pointer): Boolean;
    procedure SetFieldData(Const Dataset: TSqlitePassDataset; Const FieldBuffer: Pointer; Const DataType: TFieldType; out StrValue: String);
    function CreateTable(Const Dataset: TSqlitePassDataset): Boolean;
    procedure GetDatabaseObjects(ObjectType: TSqlitePassDatabaseObjectTypes); virtual;
    procedure GetIndexDefs; virtual;
    function IsSystemTable(TableName: String): Boolean; virtual;
  end;

TSqlitePassTranslator_Kexi = class(TSqlitePassTranslator)
  public
    procedure GetDatabaseObjects(ObjectType: TSqlitePassDatabaseObjectTypes); override;
    function IsSystemTable(TableName: String): Boolean; override;
  end;


{ ----- Blob fields stream ----- }
TSqlitePassBlobStream = class(TStream)
  private
    FField      : TBlobField;
    FFieldNo    : Integer;
    FDataSet    : TSqlitePassDataset;
    FActiveRec  : TMemRecord;
//    FCached     : Boolean;
    FMode       : TBlobStreamMode;
    FOpened     : Boolean;
    FModified   : Boolean;
    FPosition   : Longint;
    FBlobData   : TBlobData; // Not needed ?
//    FBlobSize   : Integer;
    function GetBlobSize: Longint;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    Function Read(var Buffer; Count: Longint): Longint; override;
    Function Write(const Buffer; Count: Longint): Longint; override;
    Function Seek(Offset: Longint; Origin: Word): Longint; override; // Ok
    Procedure Truncate;

  End;


 { TSqlitePassFieldDefParserList }

 TSqlitePassFieldDefParserList = Class(TSqlitePassObjectList)
  Protected
    Function GetItems(Index: Integer): TSqlitePassFieldDefParser;
    Procedure SetItems(Index: Integer; ADatabaseDefParser: TSqlitePassFieldDefParser);
  Public
    Function Add(ADatabaseDefParser: TSqlitePassFieldDefParser): Integer;
    Property Items[Index: Integer]: TSqlitePassFieldDefParser read GetItems write SetItems; default;
  end;

{ TSqlitePassBlobs }
 TSqlitePassBlobs = Class(TSqlitePassObjectList)
  Protected
    Function GetItems(Index: Integer): TSqlitePassBlobStream;
    Procedure SetItems(Index: Integer; BlobStream: TSqlitePassBlobStream);
  Public
    Function Add(BlobStream: TSqlitePassBlobStream): Integer;
    Property Items[Index: Integer]: TSqlitePassBlobStream read GetItems write SetItems; default;
  end;


 TConnectEvent  = procedure(Database: TSqlitePassDatabase) of object;

 { TVersionInfo }

 TSqlitePassVersionInfo = Class(TPersistent)
     Private
         FDatabase    : TSqlitePassDatabase;
         FComponent   : String;
         FKexi        : String;
         FSchema      : Integer;
         FSqliteLib   : String;
         FUserTag     : Integer;
         function GetFSchema: Integer;
         function GetFSqliteEngine: String;
         function GetFUserTag: Integer;
         procedure SetFComponent(const AValue: String);
         procedure SetFKexi(const AValue: String);
         procedure SetFSchema(const AValue: Integer);
         procedure SetFSqliteEngine(const AValue: String);
         procedure SetFUserTag(const AValue: Integer);
     Public
         constructor Create(Database: TSqlitePassDatabase);
     Published
         Property Component     : String  Read FComponent       Write SetFComponent;
         Property Database      : Integer Read GetFSchema       Write SetFSchema;
         Property Kexi          : String  Read FKexi            Write SetFKexi;
         Property SqliteLibrary : String  Read GetFSqliteEngine Write SetFSqliteEngine;
         Property UserTag       : Integer Read GetFUserTag      Write SetFUserTag;
     end;

TSqlitePassSynchronous = (Off, Normal, Full);
TSqlitePassTempStore = (Default, ToFile, ToMemory);
TSqlitePassEncoding = (UTF_8, UTF_16, UTF_16le, UTF_16be);
TSqlitePassPageSize = 0..8192;//512, 1024, 2048, 4096, 8192
TSqlitePassDatabaseType = (dbtSqlitePass, dbtKexi, dbtSqliteAdmin, dbtFpc, dbtUnknown);

 { TOptions }

 TSqlitePassDatabaseOptions = Class(TPersistent)
    Private
         FDatabase: TSqlitePassDatabase;
         FAutoVacuum: Boolean;
         FCacheSize: Integer;
         FDefaultCacheSize: Integer;
         FCaseSensitiveLike: Boolean;
         FCountChanges: Boolean;
         FEmptyResultCallback: Boolean;
         FEncoding: String;
         FFullColumnNames: Boolean;
         FShortColumnNames: Boolean;
          //Think to Set Also ShortColumnNames
         FPageSize: TSqlitePassPageSize;
         FSynchronous : TSqlitePassSynchronous;
         FTemporaryStorage: TSqlitePassTempStore;
         function GetFAutoVacuum: Boolean;
         function GetFCacheSize: Integer;
         function GetFCountChanges: Boolean;
         function GetFDefaultCacheSize: Integer;
         function GetFEmptyResultCallBack: Boolean;
         function GetFEncoding: TSqlitePassEncoding;
         function GetFFullColumnNames: Boolean;
         function GetFPageSize: TSqlitePassPageSize;
         function GetFSynchronous: TSqlitePassSynchronous;
         function GetFTemporaryStorage: TSqlitePassTempStore;
         procedure SetFAutoVacuum(const AValue: Boolean);
         procedure SetFCacheSize(const AValue: Integer);
         procedure SetFCaseSensitiveLike(const AValue: Boolean);
         procedure SetFCountChanges(const AValue: Boolean);
         procedure SetFDefaultCacheSize(const AValue: Integer);
         procedure SetFEmptyResultCallBack(const AValue: Boolean);
         procedure SetFEncoding(const AValue: TSqlitePassEncoding);
         procedure SetFFullColumnNames(const AValue: Boolean);
         procedure SetFPageSize(AValue: TSqlitePassPageSize);
         procedure SetFSynchronous(const AValue: TSqlitePassSynchronous);
         procedure SetFTemporaryStorage(const AValue: TSqlitePassTempStore);
    Public
         constructor Create(Database: TSqlitePassDatabase);
    Published
         Property AutoVacuum: Boolean Read GetFAutoVacuum Write SetFAutoVacuum;
         Property CacheSize: Integer Read GetFCacheSize Write SetFCacheSize;
         Property DefaultCacheSize: Integer Read GetFDefaultCacheSize Write SetFDefaultCacheSize;
         Property CaseSensitiveLike: Boolean Read FCaseSensitiveLike Write SetFCaseSensitiveLike;
         Property CountChanges: Boolean Read GetFCountChanges Write SetFCountChanges;
         Property EmptyResultCallback: Boolean Read GetFEmptyResultCallBack Write SetFEmptyResultCallBack;
         Property Encoding: TSqlitePassEncoding Read GetFEncoding Write SetFEncoding;
         Property FullColumnNames: Boolean Read GetFFullColumnNames Write SetFFullColumnNames; //Think to Set Also ShortColumnNames
         Property PageSize: TSqlitePassPageSize Read GetFPageSize Write SetFPageSize;
         Property Synchronous : TSqlitePassSynchronous Read GetFSynchronous Write SetFSynchronous;
         Property TemporaryStorage: TSqlitePassTempStore Read GetFTemporaryStorage Write SetFTemporaryStorage;
    end;
         

TSqlitePassRecordsetCache = Class(TSqlitePassObjectList)
 Private
    FCurrentRecord: TMemRecord;
    FCurrentRecordIndex: Integer;
    FHasPrimaryKey: Boolean;
    function GetFCurrentRecord: TMemRecord;
    procedure SetFCurrentRecord(const Value: TMemRecord);
  Protected
    Function GetItems(Index: Integer): TMemRecord;
    Procedure SetItems(Index: Integer; ARecord: TMemRecord);
  public
    Owner: TObject;
    MoveState: TGetResult;
    NewTempRecord: TMemRecord;
    ColumnCount: Cardinal;
    RowCount: Cardinal;
    CurrentRecordIndex: Integer;
    Procedure MoveBy(Offset: Integer);
    destructor Destroy; override;
    Property CurrentRecord: TMemRecord Read GetFCurrentRecord write SetFCurrentRecord;
    Function InitNewRecord(Fields: TFields): Integer;
    Function Add(ARecord: TMemRecord): Integer;
    Procedure AddNewRecord;
    Function Extract(Item: TMemRecord): TMemRecord;
    Function Remove(ARecord: TMemRecord): Integer;
    Function IndexOf(ARecord: TMemRecord): Integer;
    Procedure First;
    Procedure Prior;
    Procedure Next;
    Procedure Last;
    Procedure Insert(Index: Integer; ARecord: TMemRecord);
    Procedure Close;
    Procedure GetRecords(DatabaseHandle: Pointer; Sql: String; Fields: TFields = nil; SkipColumnHeader: Boolean = true; SkipFirstColumn: Boolean = true);
    Property HasPrimaryKey: Boolean Read FHasPrimaryKey;
    Property Items[Index: Integer]: TMemRecord read GetItems write SetItems; default;
    constructor Create(Owner: TObject);
  end;

 { TSqlitePassDatabase }

 TSqlitePassDatabaseAttached = Class(TObject)
       Private
         FParentDatabase     : TSqlitePassDatabase;
         FDatabase           : String;
         FName               : String;
         FDatasets           : TSqlitePassDatasets;
         FTableDefs          : TSqlitePassTableDefs;
         FIndexDefs          : TSqlitePassDatabaseIndexDefs;
         FQueryDefs          : TSqlitePassQueryDefs;
         FViews              : TSqlitePassViews;
         Procedure SetFDatabase(Value : String);
       Public
         { ----- Constructor / Destructor ----- }
         Constructor Create(AOwner : TObject);
         Destructor Destroy; override;
         { ----- Properties ----- }
         Property Database: String  Read FDatabase Write SetFDatabase;
         Property Datasets: TSqlitePassDatasets Read FDatasets Write FDatasets;
         Property IndexDefs: TSqlitePassDatabaseIndexDefs  Read FIndexDefs Write FIndexDefs;
         Property Name: String Read FName;
         property TableDefs: TSqlitePassTableDefs Read FTableDefs Write FTableDefs;
         property QueryDefs: TSqlitePassQueryDefs Read FQueryDefs Write FQueryDefs;
         property Views: TSqlitePassViews Read FViews Write FViews;
      end;

 TSqlitePassDatabase = Class(TComponent)
       Private
         FActive             : Boolean;
         FCollatingOrder     : String;
         FComponentVersion   : String;
         FDatabase           : String;
         FDatasets           : TSqlitePassDatasets;
         FTableDefs          : TSqlitePassTableDefs;
         FIndexDefs          : TSqlitePassDatabaseIndexDefs;
         FQueryDefs          : TSqlitePassQueryDefs;
         FViews              : TSqlitePassViews;
         FDatabases          : TSqlitePassDatabasesAttached;
         FDatabaseType       : TSqlitePassDatabaseType;
         FFieldDefsParser    : TSqlitePassFieldDefParserList;
         FTransactions       : TSqlitePassTransaction ;
         FOptions            : TSqlitePassDatabaseOptions;
         FQueryTimeout       : Integer;
         FReadOnly           : Boolean;
         FShowSysObjects     : Boolean;
         FSmartOpen          : Boolean;
         FUser               : Integer;
         FVersionInfo        : TSqlitePassVersionInfo;
         FInternalTransaction: Boolean;
         { ----- Events ---- }
         FAfterConnect       : TConnectEvent;
         FAfterDisconnect    : TConnectEvent;
         FBeforeConnect      : TConnectEvent;
         FBeforeDisconnect   : TConnectEvent;
         Procedure SetFDatabase(Value : String);
         procedure SetFDatabaseType(const AValue: TSqlitePassDatabaseType);
         Procedure SetFDatabaseVersion(Value : String);
         procedure SetFUser(const AValue: Integer);
         Procedure SetFVersionInfo(Value : String);
         Function  GetFCollatingOrder:String;
         Procedure SetFShowSysObjects(Value : Boolean);
         Procedure SetFReadOnly(Value : Boolean);
         Procedure SetFComponentVersion(Value: String);
         Procedure SetFActive(Value : Boolean);
         procedure ClearDefinitions;
         { ------ Pragma ------ }
         procedure setPragma(PragmaName:string; value:Boolean);overload;
         procedure setPragma(PragmaName:string; value:integer);overload;
         procedure setPragma(PragmaName, value:string);overload;
         function GetPragma(PragmaName:string):integer;
       Protected
       Public
         Engine: TSqlitePassEngine;
         Translator: TSqlitePassTranslator;
         { ----- Constructor / Destructor ----- }
         Constructor Create(AOwner : TComponent); override;
         Destructor Destroy; override;
         { ----- Procedures / Functions ----- }
         Procedure Idle;
         Procedure Open;
         Procedure Close;
         Function Compact: Integer;
         Function AttachDatabase(FAttachedDatabase, DatabaseAlias: String): Boolean;
         Function DetachDatabase(DatabaseAlias: String): Boolean;
         procedure RefreshDefinitions;
         Function IsSystemTable(TableName: String): Boolean;
         { ----- Properties ----- }
         Property FieldDefsParser: TSqlitePassFieldDefParserList Read FFieldDefsParser;
         Property Transactions: TSqlitePassTransaction Read FTransactions Write FTransactions;
       Published
         { ----- Properties ----- }
         Property Connected            : Boolean Read FActive              Write SetFActive Default False;
         Property Database             : String  Read FDatabase            Write SetFDatabase;
         Property DatabaseType         : TSqlitePassDatabaseType  Read FDatabaseType   Write SetFDatabaseType;
         Property Options              : TSqlitePassDatabaseOptions Read FOptions            Write FOptions;
         Property QueryTimeout         : Integer Read FQueryTimeout        Write FQueryTimeout;
         Property ReadOnly             : Boolean Read FReadOnly            Write SetFReadOnly;
         Property ShowSystemObjects    : Boolean Read FShowSysObjects      Write SetFShowSysObjects;
         Property SmartOpen            : Boolean Read FSmartOpen           Write FSmartOpen;
         Property UserTag              : Integer Read FUser                Write SetFUser;
         Property VersionInfo          : TSqlitePassVersionInfo  Read FVersionInfo   Write FVersionInfo;
         Property Datasets: TSqlitePassDatasets Read FDatasets Write FDatasets;
         Property IndexDefs: TSqlitePassDatabaseIndexDefs  Read FIndexDefs Write FIndexDefs;
         property TableDefs: TSqlitePassTableDefs Read FTableDefs Write FTableDefs;
         property QueryDefs: TSqlitePassQueryDefs Read FQueryDefs Write FQueryDefs;
         property Views: TSqlitePassViews Read FViews Write FViews;
         property Databases: TSqlitePassDatabasesAttached Read FDatabases;
         { ----- Events ----- }
         Property OnAfterConnect       : TConnectEvent Read FAfterConnect        Write FAfterConnect;
         Property OnAfterDisconnect    : TConnectEvent Read FAfterDisconnect     Write FAfterDisconnect;
         Property OnBeforeConnect      : TConnectEvent Read FBeforeConnect       Write FBeforeConnect;
         Property OnBeforeDisconnect   : TConnectEvent Read FBeforeDisconnect    Write FBeforeDisconnect;
      end;


TSqlitePassExportMethod        = (VisibleFields, AllFields);
TSqlitePassExportProgressEvent = procedure(Current, Total:Integer) of object;
TSqlitePassImportProgressEvent = procedure(Current:Integer) of object;


{ TSqlitePassDataset }

TSqlitePassDataset = class(TDataSet)
private
    { Hold prepared sql statements to speed up operations }
    InsertStmt:pointer;
    UpdateStmt:pointer;
    DeleteStmt:pointer;
    PreparedInsertStmt: String;
    PreparedUpdateStmt: String;
    PreparedDeleteStmt: String;
    { --------------------------------------------------- }
    FMemRecordset : TSqlitePassRecordsetCache;
    FParamCheck: Boolean;
    FParams: TParams;
    FComponentVersion       : String;
    FDatasetName            : String;
    FDatasetNameExt         : String;
    FDatasetType            : TSqlitePassDatasetType;
//    Procedure FOnGetMemoText(Sender: TField; var Text: String; DisplayText: Boolean);
    Function GetActiveRecord:  TMemRecord;
    Function FilterRecord(Buffer: PChar): Boolean;
    function RecordFilter: Boolean;
protected
    FDatabase               : TSqlitePassDatabase;
    FActive                 : Boolean;
    FReadOnly               : Boolean;
    FProcessMessages        : Boolean;
    FDetailRecordset        : TSqlitePassDataset;
    FSQL                    : TStrings;
    FSortedBy               : TStrings;
    FRefreshSorted          : Boolean;
    FMasterLink             : TMasterDataLink;
    FMasterFields           : TStrings;
    FMasterAutoActivate     : Boolean;
    FDatabaseAutoActivate   : Boolean;
    FIndexDefs              : TSqlitePassDatasetIndexDefs;
    FOnExportProgress       : TSqlitePassExportProgressEvent;
    FOnImportProgress       : TSqlitePassImportProgressEvent;
    FExportMethod           : TSqlitePassExportMethod;
    Procedure SetFComponentVersion(Value: String);
    Function  GetFDatabase:TSqlitePassDatabase;
    Procedure SetFDatabase(Value:TSqlitePassDatabase);
    Function  GetFDatasetName:String;
    Procedure SetFDatasetName(Value:String);
    Procedure SetFReadOnly(Value:Boolean);
    Procedure SetFSort(Value:TStrings);
    Procedure SetFSQL(Value:TStrings);
//    Function  GetFMasterSource: TDataSource;
//    Procedure SetFMasterSource(Value: TDataSource);
//    Procedure FProcessMasterFields(Value:TStrings);
//    Procedure SetFMasterFields(Value:TStrings);
//    Procedure SetFMaster(Value:TStrings);
//    Procedure SetFDetail(Value:TStrings);
    { ----- Filtering ----- }
//    Procedure SetFOnFilterRecord(Value: TFilterRecordEvent);
    { ----- Params ----- }
    Procedure SetParamsList(Value: TParams);
    Procedure UpdateParamsList(Sender: TObject);
    Procedure WriteParamData(Writer: TWriter);
    Function  GetParamsCount: Word;
    Procedure DefineProperties(Filer: TFiler); override;
    Procedure ReadParamData(Reader: TReader);
    { ----- Master Dataset ----- }
    (* Procedure MasterDatasetChanged;
    Procedure UpdateFromMaster;
    Procedure MasterChanged(Sender: TObject);
    Procedure MasterDisabled(Sender: TObject);
    Procedure DoOnNewRecord; override;*)
    //
    Procedure OpenRecordset;
    { ----- Override methods from TDataset ----- }
    Procedure Loaded; override;
    Procedure Notification(AComponent: TComponent; Operation: TOperation);Override;
    Function AllocRecordBuffer: PChar; override;
    Function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override; //Ok v1.00
    Function GetRecordSize: Word; override;
    Function IsCursorOpen: Boolean; override;
    Procedure ClearCalcFields(Buffer: PChar);override;
    Procedure FreeRecordBuffer(var Buffer: PChar); override;
    Procedure InternalClose; override;
    Procedure InternalInitFieldDefs; override;
    Procedure InternalInitIndexDefs;
    Procedure InternalInitRecord(Buffer: PChar); override;
    Procedure InternalOpen; override;
    Procedure SetFieldData(Field: TField; Buffer: Pointer);override;
    Procedure UpdateIndexDefs; override;
    Function  GetCanModify: Boolean; override;
    { ----- Navigation and Editing ----- }
    Procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    Procedure InternalCancel; override;
    Procedure InternalDelete; override;
    Procedure InternalEdit; override;
    Procedure InternalFirst;override;
    procedure InternalInsert; override;
    Procedure InternalLast;override;
    Procedure InternalPost; override;
    Procedure InternalRefresh; override;
    Procedure InternalSetToRecord(Buffer: PChar); override;
    { ----- Bookmark ----- }
    Procedure InternalClearBookmarks;
    Procedure InternalGotoBookmark(ABookmark: Pointer); override;
    // Procedure InternalMoveToBookmark(Bookmark: Pointer);
    Function  GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override; {From TDataset} //Ok v1.00
    Procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override; {From TDataset} //Ok v1.00
    Function  GetBookmarkStr: TBookmarkStr; override;
    Procedure SetBookmarkStr(const Value: TBookmarkStr); override;
    Procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;//Ok v1.00
    Procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;//Ok v1.00
    { ----- Miscalleous ---- }
    Procedure InternalHandleException; override;
    { ----- Records infos ----- }
    procedure AllocateBLOBPointers(Buffer: PChar);
    procedure FreeBlobPointers(Buffer: PChar);
    procedure FreeRecordPointers(Buffer: PChar);
    Function GetRecordCount  : Integer; override;
    Function GetRecNo        : Integer; override;
    Procedure SetRecNo        (Value: Integer); override;
    { ----- SQL and Utilities ----- }
 (*   Procedure StringToList(Items: String; List: TStringList);
    Function BuildKeySQL(KN,KV:TStringList):String;
    Function BuildLocateSQL(KN,KV:TStringList;Options: TLocateOptions):String;
    Function BuildDetailSQL  : String;
    Function InsertSQLString(MDString: String): String;
    Function FBindSQLParams(SQL:TStrings):String;
   // Procedure RefreshQueryParams;*)
  public
    DebugScreen: TMemo;
    Constructor Create(AOwner: TComponent); override;
    Destructor  Destroy; override;
    //  Procedure Post; override;

    Procedure Resync(Mode: TResyncMode);override;
    Procedure Reindex;
    Function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    Function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    Function BookmarkValid(Bookmark: TBookmark): Boolean; override;
    Property MasterLink: TMasterDataLink Read FMasterLink;
    Property ParamCount: Word read GetParamsCount;
    Property DatasetType: TSqlitePassDatasetType Read FDatasetType;
    //Function Requery : Boolean;
   Procedure Sort;
  published
    Property ComponentVersion: String  Read FComponentVersion Write SetFComponentVersion;
    Property Database: TSqlitePassDatabase Read GetFDatabase Write SetFDatabase;
    Property ExportMethod: TSqlitePassExportMethod Read FExportMethod Write FExportMethod;
    Property DatasetName: String Read GetFDatasetName Write SetFDatasetName;
    Property SQL: TStrings Read FSQL Write SetFSQL;
    Property Params: TParams read FParams Write SetParamsList Stored False;
    Property FieldDefs;
    Property IndexDefs: TSqlitePassDatasetIndexDefs  Read FIndexDefs Write FIndexDefs;
    Property ReadOnly: Boolean Read FReadOnly Write SetFReadOnly;
    Property DatabaseAutoActivate: Boolean Read FDatabaseAutoActivate  Write FDatabaseAutoActivate;
    Property OnExportProgress: TSqlitePassExportProgressEvent Read FOnExportProgress Write FOnExportProgress;
    Property OnImportProgress: TSqlitePassImportProgressEvent Read FOnImportProgress Write FOnImportProgress;
//        Property OnFilterRecord          : TFilterRecordEvent read FOnFilterRecord write SetFOnFilterRecord;
    Property ParamCheck: Boolean Read FParamCheck Write FParamCheck;
    Property ProcessMessages: Boolean Read FProcessMessages Write FProcessMessages;
    Property BeforeOpen;
    Property AfterOpen;
    Property BeforeClose;
    Property AfterClose;
    Property BeforeInsert;
    Property AfterInsert;
    Property BeforeEdit;
    Property AfterEdit;
    Property BeforePost;
    Property AfterPost;
    Property BeforeCancel;
    Property AfterCancel;
    Property BeforeDelete;
    Property AfterDelete;
    Property BeforeScroll;
    Property AfterScroll;
    Property OnCalcFields;
    Property OnDeleteError;
    Property OnEditError;
    Property OnNewRecord;
    Property OnPostError;
    Property AutoCalcFields;
    Property Active;
End;

implementation
Uses Dialogs, TypInfo;

{ ----- Utilities ----- }

function GetExeDir: String;
begin
 SetLength(Result,1001);
 GetModuleFileName(HInstance,PChar(Result),1000);
 Result := ExtractFilePath(StrPas(PChar(Result)));
end;

function GetWorkDir: String;
begin
 GetDir(0, Result);
 if Result[Length(Result)] <> '\' then Result:=Result+'\';
end;

function _ExecCallback(cbParam: PSqlitePassExecCallBack; Columns: Integer;
                       ColumnValues, ColumnNames: Ppchar): integer; cdecl;
begin
try
Result:=cbparam^(columns,columnValues,columnNames);
except
 result:=1;
end;
end;


{ TSqlitePassDatasets }

destructor TSqlitePassDatasets.Destroy;
begin
//  FNotifier.Free;
  inherited;
end;

Function TSqlitePassDatasets.GetItem(Index: Integer): TSqlitePassDataset;
begin
  Result:=TSqlitePassDataset(Inherited Items[Index]);
end;

Procedure TSqlitePassDatasets.SetItem(Index: Integer; Dataset: TSqlitePassDataset);
begin
  Put(Index,Dataset);
end;

function TSqlitePassDatasets.DatasetByName(Const Name: String): TSqlitePassDataset;
var
Index: Integer;
begin
Result := Nil;
for index := 0 to Count -1 do
    if Items[Index].Name = Name
       then begin
            result := Items[Index];
            Exit
            end;
If Result = nil then raise Exception.CreateFmt(Msg2070,[Name])
end;


{ TSqliteTranslator }

//-- constructor / Destructor --------------------------------------------------
constructor TSqlitePassTranslator.Create(Owner: TSqlitePassDatabase);
begin
FDatabase := Owner;
FTempDataset := TSqlitePassRecordSetCache.Create(Self);
FDateSeparator := '-';
FShortDateFormat := 'yyyy-mm-dd';
FTimeSeparator := ':';
FLongTimeFormat := 'hh:mm:ss';
FDecimalSeparator := '.';
end;

destructor TSqlitePassTranslator.Destroy;
begin
FTempDataset.Free;
end;

{ ----- TSqlitePassTranslator ----- Utilities Procedures }

Procedure TSqlitePassTranslator.SetDateFormatSettings;
begin
  FOldDateSeparator:= DateSeparator;
  FOldShortDateFormat := ShortDateFormat;
  DateSeparator := FDateSeparator;
  ShortDateFormat:= FShortDateFormat;
end;

Procedure TSqlitePassTranslator.RestoreDateFormatSettings;
begin
  DateSeparator := FOldDateSeparator;
  ShortDateFormat := FOldShortDateFormat;
end;

Procedure TSqlitePassTranslator.SetFloatFormatSettings;
begin
  FOldDecimalSeparator:= DecimalSeparator;
  DecimalSeparator := FDecimalSeparator;
end;

Procedure TSqlitePassTranslator.RestoreFloatFormatSettings;
begin
  DecimalSeparator := FOldDecimalSeparator;
end;

procedure TSqlitePassTranslator.SetTimeFormatSettings;
begin
  FOldTimeSeparator:= TimeSeparator;
  FOldLongTimeFormat := LongTimeFormat;
  TimeSeparator := FTimeSeparator;
  LongTimeFormat:= FLongTimeFormat;
end;

Procedure TSqlitePassTranslator.RestoreTimeFormatSettings;
begin
  TimeSeparator := FOldTimeSeparator;
  LongTimeFormat := FOldLongTimeFormat;
end;


{ ----- TSqlitePassTranslator ----- GetFieldData Functions }

function TSqlitePassTranslator.GetFieldData(Const Dataset: TSqlitePassDataset; Const StrValue: String; Const DataType: TFieldType; FieldBuffer: Pointer): Boolean;
begin
Try
Case DataType of
      ftBoolean    : Result := StrToBooleanFieldBuffer(StrValue, WordBool(FieldBuffer^));

      ftBytes      : Result := StrToByteFieldBuffer(StrValue, String(FieldBuffer));

      ftWord       ,
      ftSmallint   ,
      ftInteger    ,
      ftAutoInc    : Result := StrToIntFieldBuffer(StrValue, Integer(FieldBuffer^));

      ftBCD        ,
      ftCurrency   ,
      ftFloat      : Result := StrToFloatFieldBuffer(StrValue, Double(FieldBuffer^));
      ftDate       : Result := StrToDateFieldBuffer(StrValue, TDateTimeRec(FieldBuffer^));
      ftTime       : Result := StrToTimeFieldBuffer(StrValue, TDateTimeRec(FieldBuffer^));
      ftDateTime   : Result := StrToDateTimeFieldBuffer(StrValue, TDateTimeRec(FieldBuffer^));
      ftString     : Result := StrToFieldBuffer(StrValue, FieldBuffer);
      ftBlob       : Result := StrToBlobFieldBuffer(Dataset, StrValue, FieldBuffer);
   End;
Except
Result := False;
end;
end;

Function TSqlitePassTranslator.StrToBooleanFieldBuffer(Const StrValue: String; out Value: WordBool): Boolean;
var
Str: String;
R: WordBool;
Begin
Result := True;
Str := LowerCase(StrValue);
Try
R := (Pos('false', Str) = 0) and
         (Pos('no', Str)    = 0) and
         (Pos('off', Str)   = 0) and
         (Pos('0', Str)     = 0);
{$IFNDEF FPC}
// This raise a SIGENV Error while debuging...
Value := R;
{$ENDIF}

Except
 Result := False;
End;
End;

function TSqlitePassTranslator.StrToIntFieldBuffer(Const StrValue: String; out Value: Integer): Boolean;
Begin
  Result := True;
  Try
   if StrValue = ''
      then Value := 0
      else Value := StrToInt(StrValue);
  Except
   Try
     Value := Round(StrToFloat(StrValue));
   Except
     Result := False;
   End;
  End;
End;


Function TSqlitePassTranslator.StrToFloatFieldBuffer(Const StrValue: String; out Value: Double): Boolean;
Begin
 Result:=True;
 Try
 SetFloatFormatSettings;
 Try
   if StrValue = ''
      then Value := 0
      else Value := StrToFloat(StrValue);
 Except
   Result := False;
 End;
 Finally
 RestoreFloatFormatSettings;
 end;
End;

Function TSqlitePassTranslator.StrToDateFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
Begin
Result := True;
  Try
  SetDateFormatSettings;
    Try
    Value.Date := 693594 + Trunc(StrToDate(StrValue));
    Except
    Result := False;
    End;
  finally
  RestoreDateFormatSettings;
  end;
end;

function TSqlitePassTranslator.StrToTimeFieldBuffer(Const StrValue: String; out Value: TDateTimeRec): Boolean;
Begin
Result := True;
  Try
  SetTimeFormatSettings;
    Try
    Value.Time := DateTimeToTimeStamp(StrToTime(StrValue)).Time;
    Except
    Result := False;
    End;
  finally
  RestoreTimeFormatSettings;
  end;
end;

Function TSqlitePassTranslator.StrToDateTimeFieldBuffer(StrValue: String; out Value: TDateTimeRec): Boolean;
var
TimePos : Integer;
Begin
Result := True;
  Try
  SetDateFormatSettings;
  SetTimeFormatSettings;
    Try
    TimePos := Pos('T', StrValue);
    if TimePos <> 0 then StrValue[TimePos] := ' ';
    Value.DateTime := TimeStampToMsecs(DateTimeToTimeStamp(StrToDateTime(StrValue)));
    Except
    Result := False;
    End;
  finally
  RestoreDateFormatSettings;
  RestoreTimeFormatSettings;
  end;
end;

function TSqlitePassTranslator.StrToFieldBuffer(Const StrValue: String; out Value: Pointer): Boolean;
Begin
  StrLCopy(PChar(Value),PChar(StrValue),Length(StrValue));
  Result := StrValue <> '';
End;

function TSqlitePassTranslator.StrToStrFieldBuffer(const StrValue: String;
  out Value: String): Boolean;
begin
   Value := StrValue;
   Result := StrValue <> '';
end;

function TSqlitePassTranslator.StrToByteFieldBuffer(const StrValue: String;
  out Value: String): Boolean;
begin
Result := StrToStrFieldBuffer(StrValue, Value);
end;

function TSqlitePassTranslator.StrToBlobFieldBuffer(const Dataset: TSqlitePassDataset; const StrValue: String;
  out Value: Pointer): Boolean;
var
m: TMemoryStream;

  begin
Result := True;
  Try
// TODO - a Try about Blobs ...
//    m := TMemoryStream(pointer(StrToInt(StrValue)));
//    Value := Dataset.FMemRecordset.Blobs[StrToInt(StrValue)].Memory;
    Except
    Result := False;
    End;
end;


//-- SetFieldData Functions ---------------------------------------------------

procedure TSqlitePassTranslator.SetFieldData(Const Dataset: TSqlitePassDataset; Const FieldBuffer: Pointer; Const DataType: TFieldType; Out StrValue: String);
begin
 case DataType of
       ftBytes            : ByteFieldBufferToStr(PChar(FieldBuffer), StrValue);
       ftBoolean          : BooleanFieldBufferToStr(WordBool(FieldBuffer^), StrValue);

       ftWord,
       ftSmallint,
       ftInteger,
       ftAutoInc          : IntFieldBufferToStr(Integer(FieldBuffer^), StrValue);

       ftBCD,
       ftCurrency,
       ftFloat            : FloatFieldBufferToStr(Double(FieldBuffer^), StrValue);
       ftDate             : DateFieldBufferToStr(TDateTimeRec(FieldBuffer^), StrValue);
       ftTime             : TimeFieldBufferToStr(TDateTimeRec(FieldBuffer^), StrValue);
       ftDateTime         : DateTimeFieldBufferToStr(TDateTimeRec(FieldBuffer^), StrValue);
       ftString           : FieldBufferToStr (PChar(FieldBuffer), StrValue);
      end;
end;


procedure TSqlitePassTranslator.BooleanFieldBufferToStr(const Value: Boolean; out StrValue: String);
begin
If Value
   then StrValue := 'True'
   else StrValue := 'False';
end;

procedure TSqlitePassTranslator.IntFieldBufferToStr(Const Value: Integer; out StrValue: String);
Begin
  Try
   StrValue := IntToStr(Value);
  Except
   StrValue := '';
  End;
End;

procedure TSqlitePassTranslator.FloatFieldBufferToStr(Const Value: Double; out StrValue: String);
Begin
 Try
  StrValue := FloatToStr(Value);
 Except
  StrValue := '';
 End;
End;

procedure TSqlitePassTranslator.DateFieldBufferToStr(const Value: TDateTimeRec; out StrValue: String);
Begin
  Try
  SetDateFormatSettings;
    Try
    StrValue := DateToStr(Value.Date - 693594);
    Except
    StrValue := '';
    End;
  finally
  RestoreDateFormatSettings;
  end;
end;

procedure TSqlitePassTranslator.TimeFieldBufferToStr(const Value: TDateTimeRec; out StrValue: String);
Begin
  Try
  SetTimeFormatSettings;
    Try
    StrValue := TimeToStr(TimeStampToDateTime(TTimeStamp(Value)));
    Except
    StrValue := '';
    End;
  finally
  RestoreTimeFormatSettings;
  end;
end;

procedure TSqlitePassTranslator.DateTimeFieldBufferToStr(const Value: TDateTimeRec; out StrValue: String);
var
TimePos : Integer;
begin
  Try
  SetDateFormatSettings;
  SetTimeFormatSettings;
    Try
    StrValue := DateTimeToStr(TimeStampToDateTime(MsecsToTimeStamp(Value.DateTime)));
    TimePos := Pos(' ', StrValue);
    StrValue[TimePos] := 'T';
    Except
    StrValue := '';
    End;
  finally
  RestoreDateFormatSettings;
  RestoreTimeFormatSettings;
  end;
end;

procedure TSqlitePassTranslator.FieldBufferToStr(Const Value: PChar; out StrValue: String);
begin
StrValue := StrPas(Value);
end;

procedure TSqlitePassTranslator.ByteFieldBufferToStr(Const Value: PChar; out StrValue: String);
begin
StrValue := StrPas(Value);
end;

Procedure TSqlitePassTranslator.BindValues(const fst: pointer; const FieldDefs: TFieldDefs; const Rec: TMemRecord);
var
i: Integer;
begin
  for i:=0 to Pred(FieldDefs.Count)
      do BindValue(fst, FieldDefs[i], Rec.FieldsValue[i]);
end;

function TSqlitePassTranslator.BindValue(const fst: pointer; Const FieldDef: TFieldDef; Const FieldValue: String): Boolean;
var
Str : String;
Int : Integer;
WBool: WordBool;
Dble: Double;
idx:integer;
begin
if not assigned(fst) then exit;
idx := FieldDef.FieldNo;
Result := False;

case FieldDef.DataType of

 ftString, ftWideString, ftFixedChar, ftMemo,
 ftDate, ftTime, ftDateTime:
    begin
     Result := StrToStrFieldBuffer(FieldValue, Str);
     if Result
        then FDatabase.Engine.CheckResult(SqliteDbv3_bind_text(fst, idx, PChar(Str), Length(Str), SQLITE_TRANSIENT));
    end;

 ftLargeInt, ftSmallInt, ftInteger, ftWord:
    begin
    Result := StrToIntFieldBuffer(FieldValue, Int);
    If Result
       then FDatabase.Engine.CheckResult(SqliteDbv3_bind_int(fst, idx, Int));
    end;

 ftBoolean:
   begin
    Result := StrToBooleanFieldBuffer(FieldValue, WBool);
    if Result
       then FDatabase.Engine.CheckResult(SqliteDbv3_bind_int(fst, idx, 1))
   end;

 ftBcd, ftFloat, ftCurrency:
    begin
     Result := StrToFloatFieldBuffer(FieldValue, Dble);
     if Result
        then FDatabase.Engine.CheckResult(SqliteDbv3_bind_double(fst, idx, Dble));
     end;

 ftBlob, ftGraphic, ftBytes, ftVarBytes:
    begin
     Result := StrToStrFieldBuffer(FieldValue, Str);
     if Result
        then FDatabase.Engine.CheckResult(SqliteDbv3_bind_blob(fst,idx,PChar(Str),Length(Str),SQLITE_TRANSIENT));
    end;

 ftUnknown: FDatabase.Engine.CheckResult(SqliteDbv3_bind_null(fst,idx));
 else raise Exception.CreateFmt(Msg3011,[ord(FieldDef.DataType)]);
end;

if Not Result
   then FDatabase.Engine.CheckResult(SqliteDbv3_bind_null(fst,idx));

end;

function TSqlitePassTranslator.CreateTable(const Dataset: TSqlitePassDataset): Boolean;
begin
// TODO : To be implemented
end;


{ TSqlitePassDatabase }

Constructor TSqlitePassDatabase.Create(AOwner : TComponent);
begin
  Inherited Create(AOwner);
  //********************************************************************* Events
  FBeforeConnect      := Nil;
  FAfterConnect       := Nil;
  FBeforeDisconnect   := Nil;
  FAfterDisconnect    := Nil;
  //********************************************************************* Properties
  FComponentVersion := TSqlitePassDatabaseVersion;
  FActive := False;
  FReadOnly := False;
  FShowSysObjects := False;
  FSmartOpen := True;
  FFieldDefsParser := TSqlitePassFieldDefParserList.create;
  DatabaseType := dbtSqlitePass;
  FVersionInfo := TSqlitePassVersionInfo.Create(Self);
  FOptions := TSqlitePassDatabaseOptions.Create(Self);
  FQueryTimeout := 0;
  FDatasets := TSqlitePassDatasets.create(Self);
  FTableDefs := TSqlitePassTableDefs.Create(Self);
  FQueryDefs := TSqlitePassQueryDefs.Create(Self);
  FViews := TSqlitePassViews.Create(Self);
  FIndexDefs := TSqlitePassDatabaseIndexDefs.Create(Self);
  FDatabases := TSqlitePassDatabasesAttached.Create(Self);
end;

Destructor  TSqlitePassDatabase.Destroy;
begin
 If FActive then Connected := False;
 FVersionInfo.Free;
 FOptions.Free;
 FFieldDefsParser.Free;
 FTransactions.Free;
 FDatasets.Free;
 FTableDefs.Free;
 FQueryDefs.Free;
 FViews.Free;
 FIndexDefs.Free;
 FDatabases.Free;
 Inherited Destroy;
end;


procedure TSqlitePassDatabase.SetFDatabaseType(const AValue: TSqlitePassDatabaseType);
const
SqlitePassFieldTypes:array[0..21] of record
              Name:string;
              Fieldtype:TFieldType;
              end=
           ((Name:'widestring'; FieldType:ftWideString), // must be before string
            (Name:'string'; FieldType:ftString),
            (Name:'smallint'; FieldType:ftSmallInt),     // must be before integer
            (Name:'largeint'; FieldType:ftLargeint),     // must be before integer
            (Name:'integer'; Fieldtype:ftInteger),
            (Name:'word'; FieldType:ftWord),
            (Name:'float'; FieldType:ftFloat),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'curency'; FieldType:ftCurrency),
            (Name:'bcd'; FieldType:ftBCD),
            (Name:'datetime'; FieldType:ftDateTime), // must be before 'date'
            (Name:'date'; FieldType:ftDate),
            (Name:'time'; FieldType:ftTime),
            (Name:'varbytes'; FieldType:ftVarBytes),
            (Name:'bytes'; FieldType:ftBytes),
            (Name:'autoinc'; FieldType:ftAutoInc),
            (Name:'fmtmemo'; FieldType:ftFmtMemo),   // must be before memo
            (Name:'memo'; FieldType:ftMemo),
            (Name:'blob'; FieldType:ftBlob),
            (Name:'graphic'; FieldType:ftGraphic),
            (Name:'typedbinary'; FieldType:ftTypedBinary),
            (Name:'fixedchar'; FieldType:ftfixedchar));

KexiFieldTypes:array[0..12] of record
              Name:string;
              Fieldtype:TFieldType;
              end=
           ((Name:'integer'; Fieldtype:ftInteger),
            (Name:'shortinteger'; FieldType:ftSmallInt),
            (Name:'biginteger'; FieldType:ftLargeint),
            (Name:'byte'; FieldType:ftBytes),
            (Name:'double'; FieldType:ftFloat),
            (Name:'float'; FieldType:ftFloat),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'datetime'; FieldType:ftDateTime), // must be before 'date'
            (Name:'date'; FieldType:ftDate),
            (Name:'time'; FieldType:ftTime),
            (Name:'text'; FieldType:ftString),
            (Name:'clob'; FieldType:ftMemo),
            (Name:'blob'; FieldType:ftBlob));

SqliteAdminFieldTypes:array[0..11] of record
              Name:string;
              Fieldtype:TFieldType;
              end=
           ((Name:'integer'; Fieldtype:ftInteger),
            (Name:'float'; FieldType:ftFloat),
            (Name:'real'; FieldType:ftFloat),
            (Name:'numeric'; FieldType:ftFloat),
            (Name:'boolean'; FieldType:ftBoolean),
            (Name:'date'; FieldType:ftDate),
            (Name:'timestamp'; FieldType:ftDateTime), // must be before 'time'
            (Name:'time'; FieldType:ftTime),
            (Name:'nvarchar'; FieldType:ftString), // must be before 'varchar'
            (Name:'varchar'; FieldType:ftString),
            (Name:'text'; FieldType:ftString),
            (Name:'blob'; FieldType:ftBlob));// How to detect ftGraphic and ftMemo ?

var
Idx: Word;
AFieldDefParser: TSqlitePassFieldDefParser;
begin
if FDatabaseType <> AValue
   then begin
FFieldDefsParser.Clear;
FDatabaseType := AValue;
Case FDatabaseType of
     dbtSqlitePass: begin
                 for Idx := 0 to High(SqlitePassFieldTypes) do
                     begin
                     AFieldDefParser := TSqlitePassFieldDefParser.Create;
                     AFieldDefParser.Name := SqlitePassFieldTypes[Idx].Name;
                     AFieldDefParser.FieldType := SqlitePassFieldTypes[Idx].FieldType;
                     FFieldDefsParser.Add(AFieldDefParser);
                     end;
                 end;

     dbtKexi:    begin
                 for Idx := 0 to High(KexiFieldTypes) do
                     begin
                     AFieldDefParser := TSqlitePassFieldDefParser.Create;
                     AFieldDefParser.Name := KexiFieldTypes[Idx].Name;
                     AFieldDefParser.FieldType := KexiFieldTypes[Idx].FieldType;
                     FieldDefsParser.Add(AFieldDefParser);
                     end;
                 end;

     dbtSqliteAdmin:    begin
                 for Idx := 0 to High(SqliteAdminFieldTypes) do
                     begin
                     AFieldDefParser := TSqlitePassFieldDefParser.Create;
                     AFieldDefParser.Name := SqliteAdminFieldTypes[Idx].Name;
                     AFieldDefParser.FieldType := SqliteAdminFieldTypes[Idx].FieldType;
                     FieldDefsParser.Add(AFieldDefParser);
                     end;
                 end;

//   dbtFpc:
end;
end;
end;



Procedure TSqlitePassDatabase.SetFComponentVersion(Value: String);
begin
  // This property is read only
end;

Procedure TSqlitePassDatabase.SetFActive(Value : Boolean);
begin
If Value
   then Open
   else Close;
end;

Procedure TSqlitePassDatabase.Open;
begin
if FActive or  (csLoading in ComponentState) then Exit;

if (FDatabase='')
   then begin
        Raise Exception.Create(Msg1005);
        Exit;
        end;

if Assigned(FBeforeConnect) then FBeforeConnect(Self);

{ Get the database fullpath if FSmartOpen is True }
if FSmartOpen and (not FileExists(FDatabase))
  then if csDesigning in ComponentState
          then FDatabase := GetWorkDir+ExtractFileName(FDatabase)
          else FDatabase := GetExeDir+ExtractFileName(FDatabase);

if not FileExists(FDatabase)
  then Raise Exception.Create(Msg1038+CRLF+FDatabase);

{ Create the appropriate DbEngine }
Engine := TSqlitePassEngine.Create;
Engine.QueryTimeout := FQueryTimeOut;

{ Create the appropriate FieldTranslator }
// TODO : continu Translator implementation...
Case FDatabaseType of
  dbtSqlitePass  : Translator := TSqlitePassTranslator.Create(Self);
  dbtKexi        : Translator := TSqlitePassTranslator_Kexi.Create(Self);
  dbtSqliteAdmin : Translator := TSqlitePassTranslator.Create(Self);
  dbtFpc         : Translator := TSqlitePassTranslator.Create(Self);
  end;

{ Try to open the database and store the handle to the database }
Engine.OpenDatabase(FDatabase);

// TODO
{ Futur plan to add user function extension... }
(* for i:=0 to fudf.Count-1 do with TSqlitePassDatabaseTMkSqliteExtension(fudf[i]) do
if not add then
raise Exception.CreateFmt('Failed to add Extension %s',[fname]); *)

FCollatingOrder:=GetFCollatingOrder;
FActive:=True;
RefreshDefinitions;
Idle;
if Assigned(FAfterConnect) then FAfterConnect(Self);
end;


Procedure TSqlitePassDatabase.RefreshDefinitions;
var
i, j: Integer;
TempDb: TSqlitePassDatabase;
begin
  if not Connected
    then Raise Exception.Create(FDatabase+CRLF+Msg1039);
{ Get information about the database tables, queries... }
 TableDefs.Refresh;
 QueryDefs.Refresh;
 IndexDefs.Refresh;
 Views.Refresh;
 if FDatabases.count > 0 then
    begin
    TempDb := TSqlitePassDatabase.Create(Self.Owner);
    { Get information about the attached databases tables, queries... }
    with TempDb do
        begin
        for i:=0 to Pred(Self.FDatabases.count) do
        begin
        Database := Self.FDatabases[i].Database;
        Connected := True;
        TableDefs.Refresh;
        QueryDefs.Refresh;
        IndexDefs.Refresh;
        Views.Refresh;
        for j:=0 to Pred(TableDefs.Count)
            do begin
               TableDefs[j].FTableNameExt := Self.FDatabases[i].Name + '.' + TableDefs[j].TableName;
               Self.TableDefs.Add(TableDefs[j]);
               end;
        for j:=0 to Pred(QueryDefs.Count)
            do begin
               QueryDefs[j].FQueryNameExt := Self.FDatabases[i].Name + '.' + QueryDefs[j].QueryName;
               Self.QueryDefs.Add(QueryDefs[j]);
               end;

        Connected := False;
        end;
        end;
    end;    
end;

Procedure TSqlitePassDatabase.ClearDefinitions;
begin
  TableDefs.Clear;
  QueryDefs.Clear;
  IndexDefs.Clear;
  Views.Clear;
end;

Procedure TSqlitePassDatabase.Close;
begin
if FActive then
   begin
     if Assigned(FBeforeDisconnect) then FBeforeDisconnect(Self);
     //CloseDatasets;
     Engine.CloseDatabase;
     Engine.Free;
     Translator.Free;
     FActive:=False;
     if Assigned(FAfterDisconnect) then FAfterDisconnect(Self);
   end;
end;

function TSqlitePassDatabase.GetPragma(pragmaName: String): integer;
begin
if Assigned(Engine) and FActive
   then Result := Engine.GetPragma('PRAGMA ' + PragmaName)
   else Result := 0;
end;

procedure TSqlitePassDatabase.setPragma(PragmaName: string; value: Boolean);
var
 s:string;
begin
if Assigned(Engine) and FActive
   then begin
        if Value
           then s:='ON'
           else s:='OFF';
        Engine._ExecSql(Format('PRAGMA %s=%s',[PragmaName, s]));
        end;
end;

procedure TSqlitePassDatabase.setPragma(PragmaName: string; value: integer);
begin
if Assigned(Engine) and FActive
   then Engine._ExecSql(Format('PRAGMA %s=%d',[PragmaName, Value]));
end;

procedure TSqlitePassDatabase.setPragma(PragmaName, Value: string);
begin
if Assigned(Engine) and FActive
   then Engine._ExecSql(Format('PRAGMA %s=%s',[PragmaName, Value]));
end;


Procedure TSqlitePassDatabase.Idle;
begin
// TODO : To be implemented
end;

Procedure TSqlitePassDatabase.SetFDatabaseVersion(Value : String);
begin
  // This property is read only
end;

procedure TSqlitePassDatabase.SetFUser(const AValue: Integer);
begin
  if FUser=AValue then exit;
  FUser:=AValue;
end;

Procedure TSqlitePassDatabase.SetFVersionInfo(Value : String);
begin
  // This property is read only
end;


Function TSqlitePassDatabase.GetFCollatingOrder:String;
begin
// TODO : To be implemented
end;


Procedure TSqlitePassDatabase.SetFShowSysObjects(Value : Boolean);
begin
 FShowSysObjects := Value;
 if FActive then RefreshDefinitions;
end;


Procedure TSqlitePassDatabase.SetFReadOnly(Value : Boolean);
begin
 FReadOnly := Value; // ReadOnly is not supported by sqlite ?
end;



constructor TSqlitePassDatasets.Create(Owner: TSqlitePassDatabase);
begin
FDatabase := Owner;
end;


function TSqlitePassDatasets.FindDataset(
  const Value: string): TSqlitePassTrigger;
begin

end;

function TSqlitePassDatabase.IsSystemTable(TableName: String): Boolean;
begin
Result := Translator.IsSystemTable(TableName);
end;

function TSqlitePassDatabase.Compact: Integer;
var
f: TFileStream;
OriginalSize: Integer;
begin
f := TFileStream.Create(FDatabase, fmOpenRead or fmShareDenyNone);
OriginalSize := f.Size;
Engine.exec('Vacuum');
Result := OriginalSize - f.Size;
end;

function TSqlitePassDatabase.AttachDatabase(FAttachedDatabase, DatabaseAlias: String): Boolean;
var
AttachedDb: TSqlitePassDatabaseAttached;
begin
if DatabaseAlias = ''
   then begin
        DatabaseAlias := StringReplace(ExtractFileName(FAttachedDatabase), ' ', '', [rfReplaceAll]);
        DatabaseAlias := StringReplace(DatabaseAlias, '.', '_', [rfReplaceAll]);
        end;
Try
Engine.exec('ATTACH DATABASE "' + FAttachedDatabase +'" AS "'+ DatabaseAlias +'";');
AttachedDb := TSqlitePassDatabaseAttached.Create(Self);
AttachedDb.FDatabase := FAttachedDatabase;
AttachedDb.FName := DatabaseAlias;
FDatabases.Add(AttachedDb);
Finally
RefreshDefinitions;
end;
end;

function TSqlitePassDatabase.DetachDatabase(DatabaseAlias: String): Boolean;
var
AttachedDb: TSqlitePassDatabaseAttached;
begin
Try
Engine.exec('DETACH DATABASE "' + DatabaseAlias +'";');
AttachedDb := FDatabases.DatabaseByName(DatabaseAlias);
if Assigned(AttachedDb)
   then FDatabases.Remove(AttachedDb);
Finally
RefreshDefinitions;
end;
end;

procedure TSqlitePassDatabase.SetFDatabase(Value: String);
begin
  if (FActive) then
     begin
       Raise Exception.Create(Msg1010);
       Exit;
     end;
  FDatabase:=Value;

  if FSmartOpen then
   begin
   if (Pos('.kexi', FDatabase)) > 0 then SetFDatabaseType(dbtKexi)
   else if (Pos('.s3db', FDatabase)) > 0 then SetFDatabaseType(dbtSqliteAdmin)
   else if (Pos('.db', FDatabase)) > 0 then SetFDatabaseType(dbtFpc)
   else SetFDatabaseType(dbtSqlitePass);
   end;
end;

{ TSqlitePassVersionInfo }

procedure TSqlitePassVersionInfo.SetFComponent(const AValue: String);
begin
// Read only - Used to display infos in object inspector
end;

function TSqlitePassVersionInfo.GetFSqliteEngine: String;
begin
 CheckSqliteDbv3EngineLoaded;
 Result := SqliteDbv3_SqliteLibVersion;
end;

function TSqlitePassVersionInfo.GetFUserTag: Integer;
begin
 Result := FDatabase.GetPragma('user_version');
end;

function TSqlitePassVersionInfo.GetFSchema: Integer;
begin
 Result := FDatabase.GetPragma('schema_version');
end;

procedure TSqlitePassVersionInfo.SetFKexi(const AValue: String);
begin
// Read only - Used to display infos in object inspector
end;

procedure TSqlitePassVersionInfo.SetFSchema(const AValue: Integer);
begin
// Read only - Used to display infos in object inspector
end;

procedure TSqlitePassVersionInfo.SetFSqliteEngine(const AValue: String);
begin
// Read only - Used to display infos in object inspector
end;

procedure TSqlitePassVersionInfo.SetFUserTag(const AValue: Integer);
begin
  FDatabase.SetPragma('user_version', AValue);
  FUserTag := AValue;
end;

constructor TSqlitePassVersionInfo.Create(Database: TSqlitePassDatabase);
begin
inherited Create();
FDatabase := Database;
FComponent := TSqlitePassDatabaseVersion;
FKexi := '';
GetFSqliteEngine;
end;


{ TSqlitePassDatabaseOptions }

function TSqlitePassDatabaseOptions.GetFAutoVacuum: Boolean;
begin
 Result := Boolean(FDatabase.GetPragma('auto_vacuum'))
end;

function TSqlitePassDatabaseOptions.GetFCacheSize: Integer;
begin
 Result := FDatabase.GetPragma('cache_size')
end;

function TSqlitePassDatabaseOptions.GetFCountChanges: Boolean;
begin
 Result := Boolean(FDatabase.GetPragma('count_changes'));
end;

function TSqlitePassDatabaseOptions.GetFDefaultCacheSize: Integer;
begin
  Result := FDatabase.GetPragma('default_cache_size');
end;

function TSqlitePassDatabaseOptions.GetFEmptyResultCallBack: Boolean;
begin
 Result := Boolean(FDatabase.GetPragma('empty_result_callbacks'));
end;

function TSqlitePassDatabaseOptions.GetFEncoding: TSqlitePassEncoding;
begin
// TODO : To be implemented
 //Result := FDatabase.GetPragma('encoding');
 Result := UTF_8;
end;

function TSqlitePassDatabaseOptions.GetFFullColumnNames: Boolean;
begin
  Result := Boolean(FDatabase.GetPragma('full_column_names'));
end;

function TSqlitePassDatabaseOptions.GetFPageSize: TSqlitePassPageSize;
begin
  Result := FDatabase.GetPragma('page_size');
end;

function TSqlitePassDatabaseOptions.GetFSynchronous: TSqlitePassSynchronous;
begin
 Result := TSqlitePassSynchronous(FDatabase.GetPragma('synchronous'));
end;

function TSqlitePassDatabaseOptions.GetFTemporaryStorage: TSqlitePassTempStore;
begin
 Result := TSqlitePassTempStore(FDatabase.GetPragma('temp_store'));
end;

procedure TSqlitePassDatabaseOptions.SetFAutoVacuum(const AValue: Boolean);
begin
 FDatabase.SetPragma('auto_vacuum',Avalue);
 FAutoVacuum := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFCacheSize(const AValue: Integer);
begin
 FDatabase.SetPragma('cache_size',Avalue);
 FCacheSize := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFCaseSensitiveLike(const AValue: Boolean);
begin
 FDatabase.SetPragma('case_sensitive_like',Avalue);
 FCaseSensitiveLike := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFCountChanges(const AValue: Boolean);
begin
 FDatabase.SetPragma('count_changes',Avalue);
 FCountChanges := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFDefaultCacheSize(const AValue: Integer);
begin
 FDatabase.SetPragma('default_cache_size',Avalue);
 FDefaultCacheSize := AValue;
end;

procedure TSqlitePassDatabaseOptions.SetFEmptyResultCallBack(const AValue: Boolean);
begin
 FDatabase.SetPragma('empty_result_callbacks', AValue);
 FEmptyResultCallBack := aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFEncoding(const AValue: TSqlitePassEncoding);
var
s: string;
begin
Case AValue of
UTF_8: s:= 'UTF-8';
UTF_16: s:= 'UTF-16';
UTF_16le: s:= 'UTF-16le';
UTF_16be: s:= 'UTF-16be';
end;
 FDatabase.SetPragma('encoding', s);
 GetFEncoding;
end;

procedure TSqlitePassDatabaseOptions.SetFFullColumnNames(const AValue: Boolean);
begin
 FDatabase.SetPragma('full_column_names', AValue);
 FDatabase.SetPragma('short_column_names', not AValue);
 FFullColumnNames := aValue;
 FShortColumnNames := Not aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFPageSize(AValue: TSqlitePassPageSize);
begin
Case AValue of
  1..512    : AValue := 512;
  513..1024 : AValue := 1024;
  1025..2048: AValue := 2048;
  2049..4096: AValue := 4096;
  4097..8192: AValue := 8192;
  else AValue := 2048;
  end;
 FDatabase.SetPragma('page_size', AValue);
 FPageSize := aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFSynchronous(const AValue: TSqlitePassSynchronous);
begin
 FDatabase.SetPragma('synchronous', Integer(AValue));
 FSynchronous := aValue;
end;

procedure TSqlitePassDatabaseOptions.SetFTemporaryStorage(const AValue: TSqlitePassTempStore);
begin
 FDatabase.SetPragma('temp_store', Integer(AValue));
 FTemporaryStorage := aValue;
end;

constructor TSqlitePassDatabaseOptions.Create(Database: TSqlitePassDatabase);
begin
inherited Create();
FDatabase := Database;
CaseSensitiveLike := True;
end;


{ TSqlitePassFieldDefParserList }

function TSqlitePassFieldDefParserList.GetItems(Index: Integer): TSqlitePassFieldDefParser;
begin
  Result:=TSqlitePassFieldDefParser(Inherited Items[Index]);
end;

procedure TSqlitePassFieldDefParserList.SetItems(Index: Integer;
  ADatabaseDefParser: TSqlitePassFieldDefParser);
begin
   Put(Index,ADatabaseDefParser);
end;

function TSqlitePassFieldDefParserList.Add(ADatabaseDefParser: TSqlitePassFieldDefParser
  ): Integer;
begin
   Result:=Inherited Add(ADatabaseDefParser);
end;


constructor TSqlitePassDataset.Create(AOwner: TComponent);
Var
X        : Integer;
Begin
  inherited Create(AOwner);
  FComponentVersion := TSqlitePassDatasetVersion;
  FDatasetName := '';
  FMemRecordset := TSqlitePassRecordsetCache.Create(Self);
  FReadOnly := False;
  FProcessMessages := False;
  FRefreshSorted := False;
  FSQL := TStringList.Create;
  FSQL.Clear;
  FSortedBy := TStringList.Create;
  FSortedBy.Clear;
  FExportMethod := VisibleFields;
  FMasterFields := TStringList.Create;
  FIndexDefs := TSqlitePassDatasetIndexDefs.Create(Self);
  FMasterFields.Clear;
  FMasterLink := TMasterDataLink.Create(Self);
//  FMasterLink.OnMasterChange   := MasterChanged;
//  FMasterLink.OnMasterDisable  := MasterDisabled;
//  FMasterAutoActivate   := True;
  FDatabaseAutoActivate := False;
  FDatabase := Nil;
  TStringList(FSQL).OnChange := UpdateParamsList;
  FParams := TParams.Create(Self);
  FDetailRecordset := nil;
//  FOnFilterRecord   := Nil;
  FOnExportProgress := Nil;
  FOnImportProgress := Nil;
End;

destructor TSqlitePassDataset.Destroy;
Begin
  if FActive then
      Begin
        Close;
        FActive:=False;
      End;
  FMemRecordset.Free;
  FSQL.Free;
  FSortedBy.Free;
  FMasterLink.Free;
  FMasterFields.Free;
  FIndexDefs.Free;
  //  FDetail.Free;
//  FMaster.Free;
  FParams.Free;
  inherited Destroy;
End;

Procedure TSqlitePassDataset.SetFComponentVersion(Value: String);
Begin
 //*************************** ReadOnly
End;

Function TSqlitePassDataset.GetFDatabase:TSqlitePassDatabase;
Begin
 Result:=FDatabase;
End;

Procedure TSqlitePassDataset.SetFDatabase(Value:TSqlitePassDatabase);
Begin
 if Active then DatabaseError(Msg2011);
 // TODO :
 //if Assigned(FDatabase) then FDatabase.RemoveFreeNotification(Self);
 FDatabase := Value;
 FIndexDefs.FDatabase := FDatabase;
 if Assigned(FDatabase) then FDatabase.FreeNotification(Self);
End;


Function TSqlitePassDataset.GetFDatasetName:String;
Begin
 Result:= FDatasetName;
End;

Procedure TSqlitePassDataset.SetFDatasetName(Value:String);
Begin
  if Active then DatabaseError(Msg2012);
  FDatasetNameExt := Value;
  FDatasetName := Copy(Value, Pos('.',Value)+1,Length(Value));
  FDatasetType := dtUnknown;
  if Value <> ''
     then begin
      FSQL.Clear;
      if FDatabase.TableDefs.FindTable(Value) <> nil
         then FDatasetType := dtTable
         else if FDatabase.QueryDefs.FindQuery(Value) <> nil
                 then FDatasetType := dtQuery
                 else if FDatabase.Views.FindView(Value) <> nil
                      then FDatasetType := dtView;
     end;
End;

Procedure TSqlitePassDataset.SetFSQL(Value:TStrings);
Begin
 FSQL.SetText(Value.GetText);
 if Length(Value.GetText) > 0 then
    Begin
     FDatasetName:='';
     FSortedBy.Clear;
     FieldDefs.Clear;
     FIndexDefs.Clear;
    End;
End;

Procedure TSqlitePassDataset.SetFReadOnly(Value:Boolean);
Begin
 Value := Assigned(FDatabase) and (FDatabase.Connected) and (FDatabase.ReadOnly) and (not Value);
 FReadOnly:=Value;
End;


Procedure TSqlitePassDataset.SetFSort(Value:TStrings);
Begin
// TODO : To be implemented
End;

Procedure TSqlitePassDataset.Sort;
Begin
// TODO : To be implemented
End;

Procedure TSqlitePassDataset.Loaded; //Ok
begin
  try
    inherited Loaded;
  except
    Application.HandleException(Self)
  end;
end;

Procedure TSqlitePassDataset.Notification(AComponent: TComponent; Operation: TOperation); //Ok
Begin
 If (Operation = opRemove) and (AComponent = FDatabase) then FDatabase := Nil;
 Inherited Notification(AComponent, Operation);
End;

{ ----- Opening the dataset ----- }
Procedure TSqlitePassDataset.OpenRecordset;
Begin
if not Assigned(FDatabase)
   then DatabaseError(Msg2018);

if (FDatasetName='') and (SQL.Count=0)
   then DatabaseError(Msg2019);

if not (FDatabaseAutoActivate or FDatabase.Connected)
   then DatabaseError(Msg2020)
   else FDatabase.Connected:=True;

FReadOnly:= FDatabase.ReadOnly;

if FDetailRecordset <> nil
  then begin
  FDetailRecordset.Close;
  FDetailRecordset:=nil;
  end;

FDatabase.Idle;
End;

{procedure InternalOpen; virtual; abstract;

 TDataset.Open -> DoInternalOpen -> InternalOpen
------------------------------------------
Open the dataset. You must call InternalInitFieldDefs;
if DefaultFields is True, then you must call CreateFields,
which will create the necessary TFields from the fielddefs.}
Procedure TSqlitePassDataset.InternalOpen;
Var
Idx: Integer;
s: String;
Begin
Try
  if not Assigned(FDatabase)
     then DatabaseError(Msg2018);

  if (FDatasetName='') And (SQL.Count=0)
     then DatabaseError(Msg2019);

  if not FDatabaseAutoActivate
     then if FDatabase.Connected=False
             then DatabaseError(Msg2020)
     else if FDatabase.Connected=False
             then FDatabase.Connected:=True;

  if Self.Name=''
     then Self.Name:='SqlitePassDataset_'+IntToStr(Abs(GetTickCount));

  FIndexDefs.RefreshFromCache;
  Case FDatasetType of
        dtTable:   begin
                   if (FSQL.Count = 0)
                     then FSQL.Text := 'SELECT RowId, * FROM "'+FDatasetName+'";'
                     else FSQL.Text := StringReplace(Sql.Text,'SELECT','SELECT RowId',[rfReplaceAll, rfIgnoreCase]);
                     FMemRecordset.GetRecords(Database.Engine.Connection, FSQL.Text, Fields);
                   end;
        dtQuery:   begin
                   FSQL.Text := Fdatabase.QueryDefs.QueryByName(FDatasetNameExt).Sql;
                   FMemRecordset.GetRecords(Database.Engine.Connection, FSQL.Text, Fields, True, False);
                   end;
        dtView:    begin
                   FSQL.Text := Fdatabase.Views.ViewByName(FDatasetName).SqlSelectStmt;
                   FReadOnly := True;
                   FMemRecordset.GetRecords(Database.Engine.Connection, FSQL.Text, Fields, True, False);
                   end;
        dtUnknown: begin
                   FReadOnly := True;
                   FMemRecordset.GetRecords(Database.Engine.Connection, FSQL.Text, Fields, True, False);
                   end;
        end;

  InternalInitFieldDefs;


  { If persistent fields have not been created in the Fields Editor
    then FPC uses the TDataset.CreateFields procedure to create fields dynamically }
  if DefaultFields
     then CreateFields;


  { TDataset.BindFields sets up all sorts of internal
    things like field offsets, as well as creating any calculated
    or lookup fields. It is within BindFields that CalcFieldsSize
    is actually set for later use in InitBufferPointers }
  BindFields(True);

{ Prepare standard Insert, Update, Delete sql statements to speed up
engine operations. We will then just need to fill in the values using bindparams}
// Insert
s := '';
For Idx := 0 to Pred(FieldDefs.Count)
    do s := s + '?,';
System.Delete(s,Length(s),1);
PreparedInsertStmt := 'Insert into "' + FDatasetName + '" values('+ s + ');';
// Update
s := '';
for Idx := 0 to Pred(Fields.Count)
    do if (not Fields[Idx].Calculated)
          then s := s + Fields[Idx].FieldName + '= ?,';
System.Delete(s,Length(s),1);
PreparedUpdateStmt := 'Update "' + FDatasetName + '" set ' + s + ' where RowId=?;';
// Delete
PreparedDeleteStmt := 'Delete from "' + FDatasetName + '" where RowId=?;';

// everything OK: table is now open
  FActive := True;


// TODO : To be implemented
// Should Move to the first record and update the RecordCount property

except
FActive := False;
raise;
end;

End;



{procedure InternalClose; virtual; abstract;
-------------------------------------------
Closes the dataset. Any resources allocated in InternalOpen should be freed
here.}
Procedure TSqlitePassDataset.InternalClose;
Var
  I : Integer;
Begin
  BindFields(False);
  if DefaultFields
     then DestroyFields;
  FMemRecordset.Close;
  With FDatabase.Engine do
       begin
       UnprepareStmt(InsertStmt);
       UnprepareStmt(UpdateStmt);
       UnprepareStmt(DeleteStmt);
       end;

  if FDetailRecordset <> Nil then FDetailRecordset.Close;
  FDetailRecordset := nil;

// TODO : To be implemented
(*  if Assigned(FDatabase) and (not MainDatabaseShutdown) then
     Begin
     ...
     End
  Else MainDatabaseShutdown  := False;*)
End;


{procedure InternalGotoBookmark(ABookmark: Pointer); virtual; abstract;
----------------------------------------------------------------------
Set the record position on the position that is associated with the
ABookMark data. The ABookMark data is the data that is acquired through
the GetBookMarkData call, and should be kept for each record.}
Procedure TSqlitePassDataset.InternalGotoBookmark(ABookmark: Pointer);
begin
FMemRecordset.CurrentRecordIndex := PInteger(ABookmark)^;
End;

Function TSqlitePassDataset.BookmarkValid(Bookmark: TBookmark): Boolean;
Begin
Result := (PInteger(Bookmark)^ > -1) and (PInteger(Bookmark)^ < RecordCount);
End;

{ TDataset: function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; virtual; abstract;
  --------------------------------------------------------------------------
  Returns the BookmarkFlag (bfCurrent, bfBOF, bfEOF, bfInserted)
  associated with a recordBuffer.
  -------------------------------------------------------------------------- }
Function TSqlitePassDataset.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
//   TODO : Check bug index = -9998...
Try
if PInteger(Buffer)^ = NewTempRecordIndex
   then Result := bfInserted
   else Result := FMemRecordset[PInteger(Buffer)^].BookMarkFlag;
Except
   Result := bfEOF;
end;
End;

{TDataset: procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); virtual; abstract;
-----------------------------------------------------------------------------
Set the BookmarkFlag 'Value' (bfCurrent, bfBOF, bfEOF, bfInserted)
of the record Buffer.
----------------------------------------------------------------------------- }
Procedure TSqlitePassDataset.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
Begin
FMemRecordset[PInteger(Buffer)^].BookMarkFlag:= Value;
End;

function TSqlitePassDataset.GetBookmarkStr: TBookmarkStr;
begin
  Result:=inherited GetBookmarkStr;
end;

procedure TSqlitePassDataset.SetBookmarkStr(const Value: TBookmarkStr);
begin
  inherited SetBookmarkStr(Value);
end;


{procedure GetBookmarkData(Buffer: PChar; Data: Pointer); virtual; abstract;
---------------------------------------------------------------------------
Puts the bookmarkdata for Buffer into the area pointed to by Data.}
Procedure TSqlitePassDataset.GetBookmarkData(Buffer: PChar; Data: Pointer);
Begin
 if (Buffer <> Nil) and (Data <> Nil)
     then PInteger(Data)^ := FMemRecordset[PInteger(Buffer)^].RecordIndex;
End;


{procedure SetBookmarkData(Buffer: PChar; Data: Pointer); virtual; abstract;
---------------------------------------------------------------------------
Move the 'Data' value to the BookmarkData associated with Buffer}
Procedure TSqlitePassDataset.SetBookmarkData(Buffer: PChar; Data: Pointer);
Begin
 if (Buffer <> Nil) and (Data <> Nil)
   then FMemRecordset[PInteger(Buffer)^].RecordIndex := PInteger(Data)^;
End;

Procedure TSqlitePassDataset.InternalInitIndexDefs;
Begin
// TODO : To be implemented
End;

Procedure TSqlitePassDataset.UpdateIndexDefs;
Begin
 FIndexDefs.Refresh;
End;


{procedure InternalInitFieldDefs; virtual; abstract;
-----------------------------------------------------------------------
This method should be called from InternalOpen, and should
initialize FieldDef definitions for all fields in a record.
It should add these definitions to the FFielddefs object.

    ftUnknown, ftString, ftSmallint, ftInteger, ftWord,
    ftBoolean, ftFloat, ftCurrency, ftBCD, ftDate,  ftTime, ftDateTime,
    ftBytes, ftVarBytes, ftAutoInc, ftBlob, ftMemo, ftGraphic, ftFmtMemo,
    ftParadoxOle, ftDBaseOle, ftTypedBinary, ftCursor, ftFixedChar,
    ftWideString, ftLargeint, ftADT, ftArray, ftReference,
    ftDataSet, ftOraBlob, ftOraClob, ftVariant, ftInterface,
    ftIDispatch, ftGuid, ftTimeStamp, ftFMTBcd);
----------------------------------------------------------------------}
Procedure TSqlitePassDataset.InternalInitFieldDefs;

  function GetFieldDefSize(FieldDefTypeStr: String): Integer;
  var
  StartPos, EndPos: Word;
  begin
  Try
  StartPos:=pos('(',FieldDefTypeStr);
  if StartPos > 0
  then begin
       EndPos:=pos(')',FieldDefTypeStr);
       if EndPos > StartPos
          then result := Trunc(StrToFloat(Copy(FieldDefTypeStr,StartPos+1,EndPos-StartPos-1)));
               (*AFieldDef.Precision:= 2; Integer((FieldSizePrecision-Trunc(FieldSizePrecision)))*10;
               end;*)
       end;
  Except
  Result := 0;
  end;
  end;

{ Return a Unique field name, adding a number if necessary to duplicate names }
  function GetUniqueFieldName(FieldName: String): String;
  var
  i: Integer;
  NewName: String;
  begin
  i := 0;
  Result := FieldName;
  if FieldDefs.IndexOf(FieldName) <> -1
     then begin
          Repeat
          Inc(i);
          NewName := FieldName + '_' + IntToStr(i);
          Until FieldDefs.IndexOf(NewName) = -1;
          Result := NewName;
          end;
  end;

var
  FieldDefPtr: Pointer;
  FieldDefTypeStr, FieldDefNameStr : String;
  p: Pchar;
  AFieldDef : TFieldDef;
  i, idx: Integer;
begin
 FieldDefs.Clear; // TODO - already done ? Doesn't seem to be completly cleared
 p := nil;
 sqliteDbv3_prepare(FDatabase.Engine.Connection,PChar(FSql.Text),-1,FieldDefPtr, p);
 SqliteDbv3_step(FieldDefPtr);
 if FMemRecordset.HasPrimaryKey
    then i := 1
    else i := 0;
 for Idx:= i to FMemRecordSet.columnCount do
 begin
 FieldDefTypeStr:= LowerCase(StrPas(sqliteDbv3_column_decltype(FieldDefPtr,Idx)));
 FieldDefNameStr:=StrPas(sqliteDbv3_column_name(FieldDefPtr,Idx));
 { A new FieldDef is created like this : TFieldDef.Create(Self,'',ftUnknown,0,False,Count+1)
   then filled with data from Sqlite }
 AFieldDef := FieldDefs.AddFieldDef;
 AFieldDef.Name := GetUniqueFieldName(FieldDefNameStr);
 AFieldDef.DisplayName :=  AFieldDef.Name;
  // TODO
  (* if FieldDefRecord.Required = '1'
     then AFieldDef.Attributes := AFieldDef.Attributes + [faRequired];*)

  { Try to get information on datatype from Sqlite and set the corresponding datatype }
  For i:=0 to Pred(Database.FieldDefsParser.Count) do
    if Pos(Database.FieldDefsParser[i].Name, FieldDefTypeStr) <> 0
       then begin
            AFieldDef.DataType := Database.FieldDefsParser[i].FieldType;
            Break;
            end;
  //  TODO
  { Now we have to set the field size and precision }
  Case AFieldDef.DataType of
       ftString : AFieldDef.Size := GetFieldDefSize(FieldDefTypeStr);
       end;
end;
 SqliteDbv3_finalize(FieldDefPtr);
 InternalInitIndexDefs;
end;



{procedure InternalHandleException; virtual; abstract;
-----------------------------------------------------
not needed yet. Just implement an empty call.}
Procedure TSqlitePassDataset.InternalHandleException;
Begin
     Application.HandleException(Self);
End;



Function TSqlitePassDataset.GetActiveRecord:  TMemRecord;
Begin
Try
        case State of
             dsBlockRead   ,
             dsBrowse      ,

//             dsCalcFields  : Result    := CalcBuffer;
//             dsFilter      : Result    := FFilterBuffer;
             dsNewValue    ,
             dsCurValue    ,
             dsEdit        ,
             dsInsert      : Result := FMemRecordset[PInteger(ActiveBuffer)^];

(*           dsOldValue:     if FOldValue=Nil
                                then Result   := ActiveBuffer
                                else Result   := FOldValue;
             dsSetKey      :  Result   := FActiveKeyBuffer;*)
             End;
Except
Result:=Nil;
End;
End;


{function GetFieldData(Field: TField; Buffer: Pointer): Boolean; virtual; abstract;
----------------------------------------------------------------------------------
Puts the data for field Field from the active buffer into Buffer.
This is called whenever a field value is demanded by a TField, so it must be
efficient.
the TDataset class doesnt know anything about how our record buffer is organized.
The next question then becomes that if the TDataset class doesnt know about our
record structure, how it does it pull out field values from this structure?
How it does it put them back in when the user is editing a record?
The answer is that it uses the GetFieldData and SetFieldData methods which a
custom dataset class must override.
These methods are called by the Tdataset.fields when it needs to retrieve
or set a specific field value from the current record buffer.
GetFieldData is prototyped as follows:

function TDataset.GetFieldData(Field: TField; Buffer: Pointer): Boolean;

The Field parameter is the field for which the value needs to be retrieved
We send back the needed value in the buffer.}
Function TSqlitePassDataset.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  ActiveRec: TMemRecord;
begin
Try
  ActiveRec := GetActiveRecord;
  if ActiveRec = nil then exit;

// TODO : fkCalculated / fkLookup fields
  (* if (Field.FieldKind = fkCalculated) or (Field.FieldKind = fkLookup) then
    begin
      ....
    end;
  DebugScreen.Lines.Add(' > GetFieldData : Field ' + Field.DisplayName + ', CurrentRecordIndex = ' + IntToStr(ActiveRec.RecordIndex));*)
  Result := Database.Translator.GetFieldData(Self, ActiveRec.FieldsValue[Pred(Field.FieldNo)], Field.DataType, Buffer);
Except
Result := False;
end;  
End;


{procedure SetFieldData(Field: TField; Buffer: Pointer); virtual; abstract;
--------------------------------------------------------------------------
Move the data associated with Field from Buffer to the activebuffer.}
Procedure TSqlitePassDataset.SetFieldData(Field: TField; Buffer: Pointer);
var
  ActiveRec: TMemRecord;
Begin
  ActiveRec := GetActiveRecord;
  If ActiveRec = nil then exit;

// TODO : To be implemented
// for fkCalculated and fkLookup
  (*
        if (Field.FieldKind=fkCalculated) or (Field.FieldKind=fkLookup) then
           Begin
                Inc(DestinationBuffer,FStartCalc+Field.Offset);
                Boolean(DestinationBuffer[0]):=(Buffer<>nil);
                if Boolean(DestinationBuffer[0]) then CopyMemory(@DestinationBuffer[1],Buffer,Field.DataSize);
           End
        Else
          Begin
            //************************************** Field Checking and Validation
            if Field.FieldKind in [fkData, fkInternalCalc] then Field.Validate(Buffer);
            if (Field.ReadOnly) and (State <> dsSetKey) then Exit; // 19.02.2003
            //********************************************************************
*)
   Database.Translator.SetFieldData(Self, Buffer, Field.DataType, ActiveRec.FieldsValue[Pred(Field.FieldNo)]);
// if not (State in [dsInternalCalc, dsCalcFields, dsFilter, dsNewValue]) then DataEvent(deFieldChange, Longint(Field));*)
End;

{procedure InternalFirst; virtual; abstract;
-------------------------------------------
This is called when 'TDataset.First' is called;
After this method, TDataset calls 'Getrecord' asking for the next record.
'Next' will inc(CurrentRecordIndex). Finally 'GetRecord' will retrieve
Record[FMemRecordset.CurrentRecordIndex] i.e. Record[0]
or return grBOF if we are already at BOF}
Procedure TSqlitePassDataset.InternalFirst;
Begin
FMemRecordset.First;
End;

{procedure InternalLast; virtual; abstract;
------------------------------------------
This is called when 'TDataset.Last' is called;
After this method, TDataset calls 'Getrecord' asking for the prior record.
'Prior' will Dec(CurrentRecordIndex). Finally 'GetRecord' will retrieve
Record[FMemRecordset.CurrentRecordIndex] i.e. Record[Pred(RecordCount)]
or return grEOF if we are already at EOF}
Procedure TSqlitePassDataset.InternalLast;
Begin
FMemRecordset.Last;
End;


{procedure InternalSetToRecord(Buffer: PChar); virtual; abstract;
----------------------------------------------------------------
Set the current record to the record in Buffer; if bookmark data
is specified in this buffer, that data can be used to determine which
record this should be.}
Procedure TSqlitePassDataset.InternalSetToRecord(Buffer: PChar);
Begin
if Buffer=Nil then Exit;
if PInteger(Buffer)^ = -9999
   then FMemRecordset.CurrentRecordIndex := FMemRecordset.Count
   else FMemRecordset.CurrentRecordIndex := PInteger(Buffer)^;
End;

procedure TSqlitePassDataset.InternalInsert;
begin
// TODO : To be implemented
inherited internalInsert;
end;

Procedure TSqlitePassDataset.InternalEdit;
Begin
// TODO : To be implemented
inherited internalEdit;
End;

Procedure TSqlitePassDataset.InternalCancel;
Begin
// TODO : To be implemented
inherited internalCancel;
End;

{procedure InternalPost; virtual; abstract;
------------------------------------------
Post the data in the active buffer to the underlying dataset. }
Procedure TSqlitePassDataset.InternalPost;
begin
Case State of
     dsEdit   : begin
                With Database.Engine do
                  begin
                  PrepareStmt(UpdateStmt, PreparedUpdateStmt);
                  FDatabase.Translator.BindValues(UpdateStmt, FieldDefs, FMemRecordset.CurrentRecord);
                  SqliteDbv3_bind_int(UpdateStmt, Length(FMemRecordset.CurrentRecord.FieldsValue)+1, FMemRecordset.CurrentRecord.PrimaryKey);
                  SqliteDbv3_step(UpdateStmt);
//                  UnPrepareStmt(UpdateStmt); No Needed - set by TSqlitePassDataset.Destroy
                  end;
                end;


     dsInsert : begin
                 With Database.Engine do
                  begin
                  PrepareStmt(InsertStmt, PreparedInsertStmt);
                  FDatabase.Translator.BindValues(InsertStmt, FieldDefs, FMemRecordset.NewTempRecord);
                  SqliteDbv3_step(InsertStmt);
//                  UnPrepareStmt(InsertStmt); No Needed - set by TSqlitePassDataset.Destroy
                 end;
                FMemRecordset.AddNewRecord;
                end;
     end;
End;

// TODO ?
(*Procedure TSqlitePassDataset.Post;
Begin
  FInPost   := True;
  Try
    Inherited Post;
  Finally
    FInPost := False;
  End;
  If (FSortedBy.Count > 0) and (FRefreshSorted) then
     Begin
       CheckBrowseMode;
       InternalClearBookmarks;
       ClearBuffers;
//      FKexiDataset.Requery;
       ActivateBuffers;
       First;
     End;
End; *)

Procedure TSqlitePassDataset.Resync(Mode: TResyncMode);
Begin
  Inherited Resync(Mode);
End;


{procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); virtual; abstract;
---------------------------------------------------------------------------------
Adds a record to the dataset. The record's data is in Buffer and Append
indicates whether the record should be appended (True) or Inserted (False).
Note that for SQL based datasets, this has no meaning.}
Procedure TSqlitePassDataset.InternalAddRecord(Buffer: Pointer; Append: Boolean);
Begin
  if Append then
     Begin
       InternalLast;
       SetBookmarkFlag(Buffer, bfEOF);
     End;
  InternalPost;
End;


{procedure InternalDelete; virtual; abstract;
--------------------------------------------
Deletes the current Record.}
Procedure TSqlitePassDataset.InternalDelete;
Var
  Rec: TMemRecord;
Begin
Rec := GetActiveRecord;
Database.Engine.exec('Delete from "' + FDatasetName + '" where RowId = ' + IntToStr(Rec.PrimaryKey));
FMemRecordset.Remove(Rec);
End;

Procedure TSqlitePassDataset.InternalRefresh;
Begin
// TODO :
End;


{function IsCursorOpen: Boolean; virtual; abstract;
--------------------------------------------------
This function should return True if data is available, even if the dataset
is not active.}
Function TSqlitePassDataset.IsCursorOpen: Boolean;
Begin
  Result:=FActive;
End;

{function AllocRecordBuffer: PChar; virtual; abstract;
-----------------------------------------------------
Must allocate enough memory to store a complete record in the dataset.
Optionally, this buffer must contain enough memory to store bookmarkdata.
The descendent must be able to construct a bookmark from this buffer.}
Function TSqlitePassDataset.AllocRecordBuffer: PChar;
Begin
GetMem(Result, GetRecordSize);
//  FillChar(Result^, FRecordBufferSize, 0);
//  AllocateBlobPointers(Result);
End;


{procedure FreeRecordBuffer(var Buffer: PChar); virtual; abstract;
-----------------------------------------------------------------
Must free the memory allocated in the AllocRecordBuffer call.}
Procedure TSqlitePassDataset.FreeRecordBuffer(var Buffer: PChar);
Begin
//  FreeRecordPointers(Buffer);
FreeMem(Buffer, GetRecordSize);
End;


{procedure InternalInitRecord(Buffer: PChar); virtual; abstract;
---------------------------------------------------------------
This method is called to initialize a field buffer when the dataset
is put into edit or append mode. Mostly, you'll want to zero out the
buffer. We cannot just fill the buffer with 0..0 since that would overwrite our
BLOB pointers. Therefore we free the blob pointers first,
then fill the buffer with zeros, then reallocate the blob pointers}
Procedure TSqlitePassDataset.InternalInitRecord(Buffer: PChar);
begin
// TODO :

//  AllocateBlobPointers(Buffer);
//FMemRecordset.AddNewRecord;
//MyPreviousRecordIndex := FMemRecordset.CurrentRecordIndex;
PInteger(Buffer)^ := FMemRecordset.InitNewRecord(Fields);
End;


Function TSqlitePassDataset.GetCanModify: Boolean;
Begin
 //TODO : Enable writing from a query
 Result := (FActive) and (not FReadOnly) and (FDatasetType = dtTable);
End;

{function GetRecordSize: Word; virtual; abstract;
------------------------------------------------
Should return the record size; this includes ONLY the data portion
of the buffer; it excludes any bookmark or housekeeping info you may
have put in the buffer.}
//**********************************************************************
Function TSqlitePassDataset.GetRecordSize: Word;
begin
  Result := SqlitePasBookMarkSize;
End;


{function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; virtual; abstract;
-----------------------------------------------------------------------------------
This method must do 3 things:
1) Get the record data for the next/current/previous record, depending
   on the GetMode value. It should return
    grOK    if all was OK.
    grBOF   if the previous record was requested, and we are at the start.
    grEOF   if the next record was requested, and we are at the end.
    grError if an error occurred.

2) If DoCheck is True, and the result is grError, then an exception must be
    raised.

3) It should initialize bookmark data for this record with flag 'bfCurrent'
   This data can be stored in the buffer, if space was allocated for it with
   AllocRecordBuffer.
}
Function TSqlitePassDataset.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
Begin
if (State = dsInactive) or (Buffer = nil) then Exit;

  case GetMode of
    gmCurrent: FMemRecordset.MoveBy(0);
    gmNext:    FMemRecordset.Next;
    gmPrior:   FMemRecordset.Prior;
    end;
Result := FMemRecordset.MoveState;

  Case Result of
      grOk :  begin
              PInteger(Buffer)^ := FMemRecordset.CurrentRecordIndex;
// TODO :
//            ClearCalcFields(Buffer);
//            GetCalcFields(Buffer);
              end;
      grError: If DoCheck and not IsEmpty then DatabaseError(FDatasetName + ' : ' + Msg1040);
      end;
//DebugScreen.Lines.Add('GetRecord : CurrentRecordIndex = ' + IntToStr(FMemRecordset.CurrentRecordIndex));
//DebugScreen.Lines.Add('GetRecord : Dataset.state = ' + IntToStr(Ord(Self.State)));

end;

Function TSqlitePassDataset.FilterRecord(Buffer: PChar): Boolean;
var
  SaveState: TDatasetState;
Begin
// TODO :
 Result:=True;
 SaveState:=SetTempState(dsFilter);
 ClearCalcFields(Buffer);
 GetCalcFields(Buffer);
// if FRangeFiltered then Result:=FilterRange(Buffer);
// if (FFiltered) and (Result) and (Assigned(FOnFilterRecord)) then
    Begin
//      FFilterBuffer:=Buffer;
//      OnFilterRecord(Self,Result);
    End;
 RestoreState(SaveState);
End;


Procedure TSqlitePassDataset.SetRecNo(Value: Integer);
begin
CheckBrowseMode;
FMemRecordset.CurrentRecordIndex := Pred(Value);
Resync([rmCenter]);
// TODO :
//DoAfterScroll;
end;


procedure TSqlitePassDataset.InternalClearBookmarks;
begin
// TODO :
end;



//******************************************************************************
//*                  Master/Detail Handling
//******************************************************************************
(*Function  TSqlitePassDataset.GetFMasterSource : TDataSource;
Begin
 Result:= FMasterLink.DataSource;
End;

Procedure TSqlitePassDataset.SetFMasterSource(Value: TDataSource);
Begin
// TODO :
End;

Procedure TSqlitePassDataset.FProcessMasterFields(Value:TStrings);
Begin
// TODO :
End;

Procedure TSqlitePassDataset.SetFMasterFields(Value:TStrings);
Begin
// TODO :
End;

Procedure TSqlitePassDataset.SetFMaster(Value:TStrings);
Begin
 FMaster.SetText(Value.GetText);
End;

Procedure TSqlitePassDataset.SetFDetail(Value:TStrings);
Begin
 FDetail.SetText(Value.GetText);
End;


Procedure TSqlitePassDataset.MasterDatasetChanged;
Begin
// TODO :
  if csDestroying in ComponentState then EXIT;
End;


Procedure TSqlitePassDataset.RefreshQueryParams;
Begin
// TODO :
End;


Procedure TSqlitePassDataset.UpdateFromMaster;
Begin
// TODO :
End;

Procedure TSqlitePassDataset.DoOnNewRecord;
begin
// TODO :
  inherited DoOnNewRecord;
end;

Procedure TSqlitePassDataset.MasterChanged(Sender: TObject);
Begin
// TODO :
End;

Procedure TSqlitePassDataset.MasterDisabled(Sender: TObject);
Begin
 CheckBrowseMode;
// FMDisabled := not (FMasterLink.Active);
End;

*)
{ ----- Queries params ----- Begin }

Procedure TSqlitePassDataset.SetParamsList(Value: TParams);
begin
   FParams.AssignValues(Value);
end;

Procedure TSqlitePassDataset.UpdateParamsList(Sender: TObject);
var
 NewParamsList: TParams;
begin
if not (csReading in ComponentState) then
    if ParamCheck or (csDesigning in ComponentState) then
    begin
        NewParamsList := TParams.Create(Self);
        try
            NewParamsList.ParseSQL(SQL.Text, True);
            NewParamsList.AssignValues(FParams);
            FParams.Clear;
            FParams.Assign(NewParamsList);
        finally
            NewParamsList.Free;
        end;
    end;
end;

Function TSqlitePassDataset.GetParamsCount: Word;
begin
    Result := FParams.Count;
end;

Procedure TSqlitePassDataset.DefineProperties(Filer: TFiler);

    Function WriteData: Boolean;
    begin
        if Filer.Ancestor <> nil then
            Result := not FParams.IsEqual(TSqlitePassDataset(Filer.Ancestor).FParams)
        else
            Result := FParams.Count > 0;
    end;

begin
    inherited DefineProperties(Filer);
    Filer.DefineProperty('ParamData', ReadParamData, WriteParamData, WriteData);
end;

Procedure TSqlitePassDataset.ReadParamData(Reader: TReader);
begin
    Reader.ReadValue;
    Reader.ReadCollection(FParams);
end;

Procedure TSqlitePassDataset.WriteParamData(Writer: TWriter);
begin
    Writer.WriteCollection(Params);
end;

{ ----- Queries Params ----- End }

procedure TSqlitePassDataset.FreeRecordPointers(Buffer: PChar);
begin
// TODO :
  (*FreeBlobPointers(Buffer);
//  DisposeRecordID(PRecordInfo(Buffer + GetDataSize).RecordID);
  if PRecordInfo(FBookmarkStartOffset)^.BookMark <> nil then
    begin
      FreeMem(PRecordInfo(FBookmarkStartOffset)^.BookMark);
      PRecordInfo(FBookmarkStartOffset)^.BookMark := nil;
    end;*)
end;

procedure TSqlitePassDataset.AllocateBLOBPointers(Buffer: PChar);
var
  Index: Integer;
  Offset: Integer;
  Stream: TMemoryStream;
begin
// TODO :
(*  for Index := 0 to FieldCount - 1 do
    if Fields[Index].DataType in [ftMemo, ftGraphic] then
      begin
        Offset := GetFieldOffset(Fields[Index]);
        Stream := TMemoryStream.Create;
        Move(Pointer(Stream), (Buffer + Offset)^, sizeof(Pointer));
      end;*)
end;

procedure TSqlitePassDataset.FreeBlobPointers(Buffer: PChar);
var
  Index: Integer;
  Offset: Integer;
  FreeObject: TObject;
begin
// TODO :
(*  for Index := 0 to FieldCount - 1 do
    if Fields[Index].DataType in [ftMemo, ftGraphic] then
      begin
        Offset := GetFieldOffset(Fields[Index]);
        Move((Buffer + Offset)^, Pointer(FreeObject), sizeof(Pointer));
        if FreeObject <> nil then
          FreeObject.Free;
        FreeObject := nil;
        Move(Pointer(FreeObject), (Buffer + Offset)^, sizeof(Pointer));
      end; *)
end;


function TSqlitePassDataset.RecordFilter: Boolean;
var
  SaveState: TDataSetState;
begin
// TODO :
  Result := True;
  if Assigned(OnFilterRecord) then
    begin
      SaveState := SetTempState(dsFilter);
      try
//        CopyRecordToBuffer(TempBuffer);
        OnFilterRecord(Self, Result);
      except
        Application.HandleException(Self);
      end;
      RestoreState(SaveState);
    end;
end;


Procedure TSqlitePassDataset.ClearCalcFields(Buffer: PChar);
begin
//TODO
//  FillChar(Buffer[FCalcFieldsStartOffset], CalcFieldsSize, 0);
end;

function TSqlitePassDataset.GetRecordCount: Longint;
begin
   CheckActive;
   Result := FMemRecordset.Count;
end;


function TSqlitePassDataset.GetRecNo: Longint;
begin
// TODO :
//  UpdateCursorPos;
  Result := FMemRecordset.CurrentRecordIndex;
end;


Function TSqlitePassDataset.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
Begin
 Result:=TSqlitePassBlobStream.Create(TBlobField(Field),Mode);
End;


{ TSqlitePassIndex }

constructor TSqlitePassIndex.Create;
begin
inherited create;
FColumns := TSqlitePassIndexColumns.Create(Self);
end;

constructor TSqlitePassDatabaseIndex.Create(Owner: TSqlitePassDatabaseIndexDefs);
begin
inherited create;
FIndexDefs := Owner;
end;

constructor TSqlitePassDatasetIndex.Create(Owner: TSqlitePassDatasetIndexDefs);
begin
inherited create;
FIndexDefs := Owner;
end;


procedure TSqlitePassDatabaseIndex.DeleteIndex;
begin
 FindexDefs.FDatabase.Engine.Exec('Drop index ' + FIndexName);
end;

function TSqlitePassDatabaseIndex.GetAttached: Boolean;
begin
 Result := FIndexDefs.Attached;
end;

procedure TSqlitePassDatabaseIndex.Reindex;
begin
 FIndexDefs.FDatabase.Engine.Exec('Reindex ' + FIndexName);
end;

procedure TSqlitePassDatasetIndex.DeleteIndex;
begin
 FindexDefs.FDatabase.Engine.Exec('Drop index ' + FIndexName);
end;

function TSqlitePassDatasetIndex.GetAttached: Boolean;
begin
 Result := FIndexDefs.Attached;
end;

procedure TSqlitePassDatasetIndex.Reindex;
begin
 FIndexDefs.FDatabase.Engine.Exec('Reindex ' + FIndexName);
end;


destructor TSqlitePassIndex.Destroy;
begin
FColumns.Free;
inherited Destroy;
end;


{ TSqlitePassDatabaseIndexDefs }
constructor TSqlitePassDatabaseIndexDefs.Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
begin
inherited create;
FDatabase := Owner;
FAttachedDatabase := AttachedDb;
end;

procedure TSqlitePassDatabaseIndexDefs.CreateIndex(Sql: String);
begin
 FDatabase.Engine.Exec(Sql);
end;

procedure TSqlitePassDatabaseIndexDefs.CreateIndex(
  Index: TSqlitePassDatasetIndex);
var
i: integer;
SqlStmt, UniqueStmt, FieldStmt: String;
begin
If Index.Unique
   then UniqueStmt := 'UNIQUE';

SqlStmt :=  'CREATE ' + UniqueStmt + ' INDEX "' + Index.IndexName +
            '" ON "' + Index.TableName + '" (';

For i := 0 to Pred(Index.Columns.Count) do
    begin
    FieldStmt := '"'+Index.Columns[i].ColumnName+'"';
    if Index.Columns[i].Direction = cidAscending
       then FieldStmt := FieldStmt + ' ASC, '
       else FieldStmt := FieldStmt + ' DESC, ';
    SqlStmt := SqlStmt + FieldStmt;
    end;
System.Delete(SqlStmt, Length(SqlStmt)-1,1);
SqlStmt := SqlStmt + ');';
FDatabase.Engine.Exec(SqlStmt);
end;

procedure TSqlitePassDatabaseIndexDefs.DeleteIndex(IndexName: String);
begin
 FDatabase.Engine.Exec('DROP INDEX ' + IndexName + ';');
end;

function TSqlitePassDatabaseIndexDefs.FindIndex(
  const Value: string): TSqlitePassIndex;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassIndex(inherited Items[I]);
    if AnsiCompareText(Result.IndexName, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassDatabaseIndexDefs.GetAttached: Boolean;
begin
 Result := (FAttachedDatabase <> nil);
end;

function TSqlitePassDatabaseIndexDefs.GetItem(Index: Integer): TSqlitePassIndex;
begin
Result:=TSqlitePassIndex(Inherited Items[Index]);
end;

function TSqlitePassDatabaseIndexDefs.IndexByName(
  const Value: string): TSqlitePassIndex;
begin
  Result := FindIndex(Value);
  if Result = nil then
    DatabaseErrorFmt(Msg3200, [Value], FDatabase);
end;

procedure TSqlitePassDatabaseIndexDefs.Refresh;
begin
FDatabase.Translator.GetIndexDefs;
end;

procedure TSqlitePassDatabaseIndexDefs.Reindex;
begin
// TODO - Optimise
FDatabase.Engine.Exec('Reindex Binary');
end;


procedure TSqlitePassDatabaseIndexDefs.RenameIndex(OldIndexName,
  NewIndexName: String);
var
CurrentIndex: TSqlitePassIndex;
begin
if NewIndexName = ''
   then exit;
CurrentIndex := IndexByName(OldIndexName);
If Assigned(CurrentIndex)
   then begin
        CurrentIndex.IndexName := NewIndexName;
        CreateIndex(TSqlitePassDatasetIndex(CurrentIndex));
        DeleteIndex(OldIndexName);
        end;
end;

procedure TSqlitePassDatabaseIndexDefs.SetItem(Index: Integer;
  const Value: TSqlitePassIndex);
begin
  Put(Index,Value);
end;

{ TSqlitePassDatasetIndexDefs }

constructor TSqlitePassDatasetIndexDefs.Create(Owner: TSqlitePassDataset);
begin
inherited create(Owner.FDatabase);
FDatabase := Owner.FDatabase;
FDataset := Owner;
end;

procedure TSqlitePassDatasetIndexDefs.Refresh;
begin
inherited Refresh;
RefreshFromCache;
end;

procedure TSqlitePassDatasetIndexDefs.RefreshFromCache;
var
i: Integer;
TableName: String;
begin
TableName := Lowercase(FDataset.FDatasetName);
Self.Clear;
for i := 0 to Pred(FDatabase.IndexDefs.count)
 do begin
    if FDatabase.IndexDefs[i].TableName = TableName
       then Self.Add(FDatabase.IndexDefs[i]);
    end;
end;


{ TSqlitePassTableDefs }

constructor TSqlitePassTableDefs.Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
begin
FDatabase := Owner;
FAttachedDatabase := AttachedDb;
end;

procedure TSqlitePassTableDefs.CreateTable(Sql: String);
begin
 FDatabase.Engine.Exec(Sql);
end;

procedure TSqlitePassTableDefs.DeleteTable(TableName: String; Silent: Boolean = False);
var
SilentMode: String;
begin
 If Silent
    then SilentMode := 'if exists '
    else SilentMode := '';

 FDatabase.Engine.Exec('Drop table ' + SilentMode + TableName + ';');
end;

destructor TSqlitePassTableDefs.Destroy;
begin

end;

function TSqlitePassTableDefs.FindTable(
  const Value: string): TSqlitePassTableDef;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassTableDef(inherited Items[I]);
    if AnsiCompareText(Result.TableNameExt, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassTableDefs.GetAttached: Boolean;
begin
 Result := (FAttachedDatabase <> nil);
end;

function TSqlitePassTableDefs.GetItem(Index: Integer): TSqlitePassTableDef;
begin
Result:=TSqlitePassTableDef(Inherited Items[Index]);
end;

procedure TSqlitePassTableDefs.Refresh;
var
i: integer;
t: TSqlitePassTableDef;
begin
Clear;
FDatabase.Translator.GetDatabaseObjects(dboTableDefs);
for i := 0 to Pred(FDatabase.Translator.FTempDataset.count)
 do begin
    t := TSqlitePassTableDef.Create(Self);
    t.FTableName := FDatabase.Translator.FTempDataset[i].FieldsValue[2];
    t.FTableNameExt := t.FTableName; // Is modified by TSqlitePassDatabase.RefreshDefinitions
    t.FSql := FDatabase.Translator.FTempDataset[i].FieldsValue[4];
    Self.Add(t);
    end;
end;

procedure TSqlitePassTableDefs.SetItem(Index: Integer;
  const Value: TSqlitePassTableDef);
begin
  Put(Index,Value);
end;

function TSqlitePassTableDefs.TableByName(
  const Value: string): TSqlitePassTableDef;
begin
  Result := FindTable(Value);
  if Result = nil then
    DatabaseErrorFmt(Msg3400, [Value], FDatabase);
end;

{ TSqlitePassTableDef }

constructor TSqlitePassTableDef.Create(Owner: TSqlitePassTableDefs);
begin
FTableDefs := Owner;
FFieldDefs := TFieldDefs.Create(nil);
end;


procedure TSqlitePassTableDef.DeleteTable(Silent: Boolean = False);
begin
 FTableDefs.DeleteTable(FTableName, Silent);
end;

destructor TSqlitePassTableDef.Destroy;
begin
FFieldDefs.Free;
end;

procedure TSqlitePassTableDef.Reindex;
begin
//
end;

procedure TSqlitePassTableDef.RenameAs(NewName: String);
begin
//
end;


procedure TSqlitePassDataset.Reindex;
begin
 if DatasetType = dtTable
    then FDatabase.Engine.Exec('Reindex ' + FDatasetName + ';');
end;

{ TSqlitePassView }

constructor TSqlitePassView.Create(Owner: TSqlitePassViews);
begin
 FViews := Owner;
end;


procedure TSqlitePassView.DeleteView;
begin
 FViews.DeleteView(FViewName);
end;

destructor TSqlitePassView.Destroy;
begin
 inherited;
end;

function TSqlitePassView.GetAttached: Boolean;
begin
 Result := FViews.Attached;
end;

{ TSqlitePassViews }

constructor TSqlitePassViews.Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
begin
 FDatabase := Owner;
 FAttachedDatabase := AttachedDb;
end;

procedure TSqlitePassViews.CreateView(Sql: String);
begin
  FDatabase.Engine.Exec(Sql);
end;

procedure TSqlitePassViews.DeleteView(ViewName: String);
begin
 FDatabase.Engine.Exec('Drop view ' + ViewName + ';');
end;

function TSqlitePassViews.FindView(const Value: string): TSqlitePassView;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassView(inherited Items[I]);
    if AnsiCompareText(Result.ViewName, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassViews.GetAttached: Boolean;
begin
 Result := (FAttachedDatabase <> nil);
end;

function TSqlitePassViews.GetItem(Index: Integer): TSqlitePassView;
begin
 Result:=TSqlitePassView(Inherited Items[Index]);
end;

procedure TSqlitePassViews.Refresh;
var
CreateStmtStart, i: integer;
v: TSqlitePassView;
CreateStmt: String;
begin
Clear;
FDatabase.Translator.GetDatabaseObjects(dboViews);
for i := 0 to Pred(FDatabase.Translator.FTempDataset.count)
 do begin
    v := TSqlitePassView.Create(Self);
    v.FViewName := FDatabase.Translator.FTempDataset[i].FieldsValue[2];
    v.FSqlCreateStmt := FDatabase.Translator.FTempDataset[i].FieldsValue[4];
    //TODO Improve 'Select' catching
    CreateStmtStart := Pos('Select ', v.FSqlCreateStmt);
    if CreateStmtStart <> 0
       then v.FSqlSelectStmt := Copy(v.FSqlCreateStmt, CreateStmtStart, Length(v.FSqlCreateStmt));
    Self.Add(v);
    end;
end;

procedure TSqlitePassViews.SetItem(Index: Integer;
  const Value: TSqlitePassView);
begin
  Put(Index,Value);
end;

function TSqlitePassViews.ViewByName(const Value: string): TSqlitePassView;
begin
  Result := FindView(Value);
  if Result = nil then
    DatabaseErrorFmt(Msg3400, [Value], FDatabase);
end;

{ TSqlitePassTrigger }

constructor TSqlitePassTrigger.Create(Owner: TSqlitePassTriggers);
begin
 FTriggers := Owner;
end;


procedure TSqlitePassTrigger.DeleteTrigger;
begin
 FTriggers.FDatabase.Engine.Exec('Drop trigger '+ FTriggerName + ';');
end;

destructor TSqlitePassTrigger.Destroy;
begin
 inherited;
end;

function TSqlitePassTrigger.GetAttached: Boolean;
begin
 Result := FTriggers.Attached;
end;

{ TSqlitePassTriggers }

constructor TSqlitePassTriggers.Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
begin
 FDatabase := Owner;
 FAttachedDatabase := AttachedDb;
end;

procedure TSqlitePassTriggers.CreateTrigger(Sql: String);
begin
 FDatabase.Engine.Exec(Sql);
end;

procedure TSqlitePassTriggers.DeleteTrigger(TriggerName: String);
begin
  FDatabase.Engine.Exec('Drop trigger ' + TriggerName + ';');
end;


function TSqlitePassTriggers.FindTrigger(
  const Value: string): TSqlitePassTrigger;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassTrigger(inherited Items[I]);
    if AnsiCompareText(Result.TriggerName, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassTriggers.GetAttached: Boolean;
begin
 Result := (FAttachedDatabase <> nil);
end;

function TSqlitePassTriggers.GetItem(Index: Integer): TSqlitePassTrigger;
begin
Result:=TSqlitePassTrigger(Inherited Items[Index]);
end;


procedure TSqlitePassTriggers.Refresh;
begin
// TODO
end;

procedure TSqlitePassTriggers.SetItem(Index: Integer;
  Value: TSqlitePassTrigger);
begin
   Put(Index,Value);
end;

function TSqlitePassTriggers.TriggerByName(
  const Value: string): TSqlitePassTrigger;
begin
  Result := FindTrigger(Value);
  if Result = nil then
    DatabaseErrorFmt(Msg3500, [Value], FDatabase);
end;


{ TBlobStream }

Constructor TSqlitePassBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
Begin
 FMode     := Mode;
 FField    := Field;
 FDataSet  := TSqlitePassDataset(FField.DataSet);
 FPosition := 0;
 FOpened   := True;
 FActiveRec   := FDataSet.GetActiveRecord;

 if FActiveRec = Nil then Exit;

 if FDataSet.State = dsFilter
     then DatabaseErrorFmt(Msg2001, [FField.DisplayName], FDataSet);

 FFieldNo := Pred(FField.FieldNo);

 if Mode = bmWrite then Truncate;
End;

Destructor TSqlitePassBlobStream.Destroy;
Begin
 if FModified then
   try
     FField.Modified := True;
     FDataSet.DataEvent(deFieldChange, Longint(FField));
     FActiveRec   := Nil;
     FOpened   := False;
   Except
     Application.HandleException(Self);
   End;
End;

function TSqlitePassBlobStream.Read(var Buffer; Count: Longint): Longint;
begin
  Result := 0;
  if FOpened
     then begin
     if Count > Size - FPosition
        then Result := Size - FPosition
        else Result := Count;
     if Result > 0
        then begin
          Move(PChar(FActiveRec.FieldsValue[FFieldNo])[FPosition], Buffer, Result);
          FBlobData :=  FActiveRec.FieldsValue[FFieldNo]; // Remove ?
          Inc(FPosition, Result);
        end;
     end;
end;


Function TSqlitePassBlobStream.Write(const Buffer; Count: Longint): Longint;
Begin
 Result := 0;
 if FOpened then
    Begin
    Result := Count;
    SetLength(FActiveRec.FieldsValue[FFieldNo], (FPosition + Result));
    Move(Buffer, PChar(FActiveRec.FieldsValue[FFieldNo])[FPosition], Result);
    FBlobData :=  FActiveRec.FieldsValue[FFieldNo];
    Inc(FPosition, Result);
    FModified := True;
   End;
end;

Function TSqlitePassBlobStream.Seek(Offset: Longint; Origin: Word): Longint;
Begin
 Case Origin of
      soFromBeginning: FPosition := Offset;
      soFromCurrent: Inc(FPosition, Offset);
      soFromEnd: FPosition := GetBlobSize + Offset;
 End;
 Result := FPosition;
End;


Procedure TSqlitePassBlobStream.Truncate;
Begin
 if FOpened then
    Begin
     SetLength(FActiveRec.FieldsValue[FFieldNo], FPosition);
     FModified := True;
   End;
End;

function TSqlitePassBlobStream.GetBlobSize: Longint;
begin
  Result := 0;
  if FOpened and Assigned(FActiveRec)
     then Result := Length(FActiveRec.FieldsValue[FFieldNo]);
end;

{ TSqlitePassRecordsetCache }

Function TSqlitePassRecordsetCache.Add(ARecord: TMemRecord): Integer;
begin
  Result:=Inherited Add(ARecord);
end;

destructor TSqlitePassRecordsetCache.Destroy;
begin
  inherited;
end;

Function TSqlitePassRecordsetCache.Extract(Item: TMemRecord): TMemRecord;
begin
  Result:=TMemRecord(Inherited Extract(Item));
end;

Procedure TSqlitePassRecordsetCache.First;
begin
CurrentRecordIndex := -1;
end;

Procedure TSqlitePassRecordsetCache.Next;
begin
MoveBy(1)
end;

Procedure TSqlitePassRecordsetCache.Prior;
begin
MoveBy(-1)
end;

Procedure TSqlitePassRecordsetCache.Last;
begin
CurrentRecordIndex := Count;
end;

Procedure TSqlitePassRecordsetCache.MoveBy(Offset: Integer);
begin
MoveState := grOk;
if Count = 0
   then begin
        MoveState := grError;
        Exit;
        end;

Case Offset of
   { Move to Previous record }
   -1 : if CurrentRecordIndex < 1
           then MoveState := grBOF
           else Dec(CurrentRecordIndex);
   { Current record }
    0 : if (CurrentRecordIndex < 0) or (CurrentRecordIndex > Pred(Count))
           then MoveState := grError;
    { Move to the Next record }
    1 : if (CurrentRecordIndex >= Pred(Count))
           then MoveState := grEOF
           else Inc(CurrentRecordIndex);
    end;
end;

Function TSqlitePassRecordsetCache.GetItems(Index: Integer): TMemRecord;
begin
Result := nil;
if Index = NewTempRecordIndex
   then Result := NewTempRecord
   else if (Index > -1) and (Index < Count)
           then Result:=TMemRecord(Inherited Items[Index]);
end;


Function TSqlitePassRecordsetCache.IndexOf(ARecord: TMemRecord): Integer;
begin
  Result:=Inherited IndexOf(ARecord);
end;

Procedure TSqlitePassRecordsetCache.Insert(Index: Integer; ARecord: TMemRecord);
begin
  Inherited Insert(Index,ARecord)
end;

procedure TSqlitePassRecordsetCache.Close;
begin
Clear;
If Assigned(NewTempRecord)
   then begin
   NewTempRecord := nil;
   NewTempRecord.Free;
   end;
end;

Procedure TSqlitePassRecordsetCache.GetRecords(DatabaseHandle: Pointer; Sql: String; Fields: TFields = nil; SkipColumnHeader: Boolean = True; SkipFirstColumn: Boolean = True);
var
//Db: TSqlitePassDatabase;
PDataArray: Pointer;
ColIdx, RowIdx : Cardinal;
RecordOffset: PPChar;
Rec: TMemRecord;
ErrMsg   : PChar;
//OldScreenCursor : Integer;
//BlobStream: TSqlitePassBlobStream;
begin
  Close;
(*  if Owner is TSqlitePassDataset
     then Db := TSqlitePassDataset(Owner).FDatabase;
  else if Owner is TSqlitePassDatabase
      then Db := TSqlitePassDatabase(Owner)
  else if Owner is TSqlitePassEngine
      then Db := TSqlitePassEngine(Owner)


     then Db.Engine.StartChrono;*)
//  OldScreenCursor := Screen.cursor;
//  Screen.Cursor := crHourglass;

  SqliteDbv3_get_table(DatabaseHandle, PChar(Sql), PDataArray, RowCount, ColumnCount, ErrMsg);

  if StrPas(ErrMsg) <> ''
     then raise Exception.Create(StrPas(ErrMsg));

  if Assigned(PDataArray) and (RowCount > 0) then
  begin
  RecordOffset := PDataArray;
  { Skip the first record holding the Columns Headers) }
  if SkipColumnHeader
     then Inc(RecordOffset, ColumnCount);
  { Exclude the first column (PrimaryKey) }
  FHasPrimaryKey := SkipFirstColumn;
  if SkipFirstColumn
     then Dec(ColumnCount);

  { Proceed with every records }
  for RowIdx := 0 to Pred(RowCount) do
      begin
      Rec:=TMemRecord.Create(ColumnCount);
      if SkipFirstColumn
         then begin
              { We get the PrimaryKey value }
              Rec.PrimaryKey := StrToInt(StrPas(RecordOffset^));
              Inc(RecordOffset);
              end;
      { We get the fields values }
      for ColIdx := 0 to Pred(ColumnCount) do
          begin
          Rec.FieldsValue[ColIdx] := String(RecordOffset^);
          Rec.RecordIndex := RowIdx;
          Inc(RecordOffset);
          end;
          Add(Rec);
      end;
  end;
  { Create a temp record to speed up edit/insert operation }
  NewTempRecord := TMemRecord.create(ColumnCount);
  if Assigned(PDataArray)
     then SqliteDbv3_free_table(PDataArray);

(*  if Assigned(Db)
     then Db.Engine.StopChrono;*)

//  Screen.Cursor := OldScreenCursor;
end;


Function TSqlitePassRecordsetCache.Remove(ARecord: TMemRecord): Integer;
begin
  Result:=Inherited Remove(ARecord);
end;

Procedure TSqlitePassRecordsetCache.SetItems(Index: Integer; ARecord: TMemRecord);
begin
  Put(Index,ARecord);
end;

constructor TSqlitePassRecordsetCache.Create(Owner: TObject);
begin
  inherited Create;
  Self.Owner := Owner;
  First;
end;

procedure TSqlitePassRecordsetCache.SetFCurrentRecord(const Value: TMemRecord);
begin
  FCurrentRecord := Value;
end;

function TSqlitePassRecordsetCache.GetFCurrentRecord: TMemRecord;
begin
 Result := GetItems(CurrentRecordIndex);
end;

// TODO :
(*procedure TKexiDataset.OpenCursor(InfoQuery:Boolean);
begin
  try
  if fDatabaseName = .. then
  DatabaseError(SDatabaseNameMissing);
  inherited OpenCursor(InfoQuery);
  except
  raise;
  end;
end;*)

function TSqlitePassRecordsetCache.InitNewRecord(Fields: TFields) : Integer;
var
Idx: Integer;
begin
For Idx := 0 to Pred(Fields.Count)
    do NewTempRecord.fieldsValue[Fields[Idx].FieldNo-1] := Fields[Idx].DefaultExpression;
Result := NewTempRecordIndex;
end;


procedure TSqlitePassRecordsetCache.AddNewRecord;
Var
NewRec: TMemRecord;
begin
NewRec := TMemRecord.Create(ColumnCount);
NewRec.Assign(NewTempRecord);
NewRec.RecordIndex := Count;
Add(NewRec);
end;

{ TSqlitePassBlobs }

function TSqlitePassBlobs.Add(BlobStream: TSqlitePassBlobStream): Integer;
begin
  Result:=Inherited Add(BlobStream);
end;

function TSqlitePassBlobs.GetItems(Index: Integer): TSqlitePassBlobStream;
begin
   Result:=TSqlitePassBlobStream(Inherited Items[Index]);
end;

procedure TSqlitePassBlobs.SetItems(Index: Integer;
  BlobStream: TSqlitePassBlobStream);
begin
   Put(Index,BlobStream);
end;


{ TSqlitePassQueryDef }

constructor TSqlitePassQueryDef.Create(Owner: TSqlitePassQueryDefs);
begin
FQueryDefs:=Owner;
FFieldDefs := TFieldDefs.Create(nil);
end;

procedure TSqlitePassQueryDef.DeleteQuery;
begin
// TODO : To be implemented
end;

destructor TSqlitePassQueryDef.Destroy;
begin
FFieldDefs.Free;
end;

function TSqlitePassQueryDef.GetAttached: Boolean;
begin
 Result := FQueryDefs.Attached;
end;

procedure TSqlitePassQueryDef.RenameAs(NewName: String);
begin
// TODO : To be implemented
end;

{ TSqlitePassQueryDefs }

constructor TSqlitePassQueryDefs.Create(Owner: TSqlitePassDatabase; AttachedDb: TSqlitePassDatabaseAttached = nil);
begin
FDatabase := Owner;
FAttachedDatabase := nil;
end;

procedure TSqlitePassQueryDefs.CreateQuery(Sql: String);
begin
// TODO : To be implemented
end;

procedure TSqlitePassQueryDefs.DeleteQuery(QueryName: String);
begin
// TODO : To be implemented
end;

function TSqlitePassQueryDefs.FindQuery(
  const Value: string): TSqlitePassQueryDef;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassQueryDef(inherited Items[I]);
    if AnsiCompareText(Result.QueryNameExt, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassQueryDefs.GetAttached: Boolean;
begin
 Result := (FAttachedDatabase <> nil);
end;

function TSqlitePassQueryDefs.GetItem(Index: Integer): TSqlitePassQueryDef;
begin
Result:=TSqlitePassQueryDef(Inherited Items[Index]);
end;

function TSqlitePassQueryDefs.QueryByName(
  const Value: string): TSqlitePassQueryDef;
begin
  Result := FindQuery(Value);
  if Result = nil then
    DatabaseErrorFmt(Msg3300, [Value], FDatabase);
end;


procedure TSqlitePassQueryDefs.Refresh;
var
i: integer;
q: TSqlitePassQueryDef;
begin
Clear;
FDatabase.Translator.GetDatabaseObjects(dboQueryDefs);
for i := 0 to Pred(FDatabase.Translator.FTempDataset.count)
 do begin
    q := TSqlitePassQueryDef.Create(Self);
    q.FQueryName := FDatabase.Translator.FTempDataset[i].FieldsValue[1];
    q.FQueryNameExt := q.FQueryName; // Is modified by TSqlitePassDatabase.RefreshDefinitions
    q.FSql := FDatabase.Translator.FTempDataset[i].FieldsValue[2];
    Self.Add(q);
    end;
end;

procedure TSqlitePassQueryDefs.SetItem(Index: Integer;
  const Value: TSqlitePassQueryDef);
begin
  Put(Index,Value);
end;

procedure TSqlitePassTranslator.GetDatabaseObjects(
  ObjectType: TSqlitePassDatabaseObjectTypes);
var
ObjectTypeStr, QuerySql: String;
begin
if FDatabase.FActive
   then begin
   Case ObjectType of
    dboAll: ObjectTypeStr := '';
    dboTableDefs: ObjectTypeStr := ' WHERE type="table"';
    dboIndexDefs: ObjectTypeStr := ' WHERE type="index"';
    dboViews: ObjectTypeStr     := ' WHERE type="view"';
    dboTriggers: ObjectTypeStr  := ' WHERE type="trigger"';
    end;
    if ObjectType = dboQueryDefs
       then begin
            // TODO : to be implemented
            QuerySql := '';
            Exit;
            end
            else QuerySql := 'SELECT * FROM sqlite_master'+ObjectTypeStr+' ORDER BY type, tbl_name'
  end;
  FTempDataset.GetRecords(FDatabase.Engine.Connection, PChar(QuerySql), nil, True, False);
end;

procedure TSqlitePassTranslator.GetIndexDefs;
var
i, j, SubStringBegin, SubStringEnd: integer;
DirectionStr: String;
Idx: TSqlitePassIndex;
Col: TSqlitePassIndexColumn;
begin
//TODO should be optimised !

FDatabase.IndexDefs.Clear;
{ We get info on IndexName, DatasetName and SQLCreateStmt }
FTempDataset.GetRecords(FDatabase.Engine.Connection,
  PChar('SELECT * FROM sqlite_master WHERE type="index" ORDER BY type, tbl_name'), nil, True, False);
for i := 0 to Pred(FTempDataset.count)
 do begin
    Idx := TSqlitePassDatabaseIndex.Create(FDatabase.IndexDefs);
    Idx.FIndexName := FTempDataset[i].FieldsValue[1];
    Idx.FDatasetName := FTempDataset[i].FieldsValue[2];
    Idx.FSql := FTempDataset[i].FieldsValue[4];
    FDatabase.IndexDefs.Add(Idx);
    end;
{ For each table, we check if the index is unique or not }
for i := 0 to Pred(FDatabase.TableDefs.Count)
    do begin
FTempDataset.GetRecords(FDatabase.Engine.Connection,
  PChar('PRAGMA index_list ("' + FDatabase.TableDefs[i].TableName +'");' ), nil, True, False);
for j := 0 to Pred(FTempDataset.count)
 do begin
    Idx := FDatabase.IndexDefs.FindIndex(FTempDataset[j].FieldsValue[1]);
    if Assigned(Idx)
       then Idx.Unique := (FTempDataset[j].FieldsValue[2] = '1');
    end;
    end;
{ For each index, we get info about fields used and sort order }
for i := 0 to Pred(FDatabase.IndexDefs.Count)
    do begin
FTempDataset.GetRecords(FDatabase.Engine.Connection,
  PChar('PRAGMA index_info ("' + FDatabase.IndexDefs[i].IndexName +'");' ), nil, True, False);
for j := 0 to Pred(FTempDataset.count)
    do begin
    Col := TSqlitePassIndexColumn.Create;
    Col.Position := StrToInt(FTempDataset[j].FieldsValue[1]);
    Col.ColumnName := FTempDataset[j].FieldsValue[2];
    SubStringBegin := Pos('"'+Col.ColumnName+'"',FDatabase.IndexDefs[i].Sql);
    If SubStringBegin <> -1
       then begin
       DirectionStr := Copy(FDatabase.IndexDefs[i].Sql,SubStringBegin + 2 + length(Col.ColumnName), 6);
       If Pos('ASC',DirectionStr) <> 0
          then Col.Direction := cidAscending
          else if Pos('DESC',DirectionStr) <> 0
                  then Col.Direction := cidDescending
                  else Col.Direction := cidUnknown;

             end;
    FDatabase.IndexDefs[i].Columns.Add(Col);
    end;
    end;

end;

function TSqlitePassTranslator.IsSystemTable(TableName: String): Boolean;
begin
Result := False;
end;

{ TSqlitePassTranslator_Kexi }

procedure TSqlitePassTranslator_Kexi.GetDatabaseObjects(
  ObjectType: TSqlitePassDatabaseObjectTypes);
var
ObjectTypeStr, QuerySql: String;
begin
if FDatabase.FActive
   then begin
   Case ObjectType of
    dboAll: ObjectTypeStr := '';
    dboTableDefs: ObjectTypeStr := ' WHERE type="table"';
    dboIndexDefs: ObjectTypeStr := ' WHERE type="index"';
    dboViews: ObjectTypeStr     := ' WHERE type="view"';
    dboTriggers: ObjectTypeStr  := ' WHERE type="trigger"';
    end;
    if ObjectType = dboQueryDefs
       then { Special code for Kexi }
            QuerySql := 'SELECT K1.o_type,K1.o_name,K2.o_data '+
                        'FROM kexi__objects K1, kexi__objectdata K2 '+
                        'WHERE (K1.o_id=K2.o_id) AND (K1.o_type=2) AND (K2.o_sub_id="sql");'
       else { standard call for others objects }
            QuerySql := 'SELECT * FROM sqlite_master'+ObjectTypeStr+' ORDER BY type, tbl_name';
  end;
  FTempDataset.GetRecords(FDatabase.Engine.Connection, PChar(QuerySql), nil, True, False);
end;


function TSqlitePassTranslator_Kexi.IsSystemTable(
  TableName: String): Boolean;
var
i: Integer;
begin
for i := 0 to High(KexiSystemTables) do
    begin
    Result :=  (TableName = KexiSystemTables[i]);
    if Result then break;
    end;
end;

{ TSqlitePassIndexColumns }

function TSqlitePassIndexColumns.ColumnByName(
  const Value: string): TSqlitePassIndexColumn;
begin
  Result := FindColumn(Value);
  if Result = nil
     then DatabaseErrorFmt(Msg3200, [Value], nil);
end;

constructor TSqlitePassIndexColumns.Create(Owner: TSqlitePassIndex);
begin
inherited create;
FIndex := Owner;
end;

function TSqlitePassIndexColumns.FindColumn(
  const Value: string): TSqlitePassIndexColumn;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassIndexColumn(inherited Items[I]);
    if AnsiCompareText(Result.ColumnName, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassIndexColumns.GetItem(
  Index: Integer): TSqlitePassIndexColumn;
begin
Result:=TSqlitePassIndexColumn(Inherited Items[Index]);
end;

procedure TSqlitePassIndexColumns.SetItem(Index: Integer;
  const Value: TSqlitePassIndexColumn);
begin
  Put(Index,Value);
end;

{ TSqlitePassDatabasesAttached }

constructor TSqlitePassDatabasesAttached.Create(
  Owner: TSqlitePassDatabase);
begin
 FDatabase := Owner;
end;

function TSqlitePassDatabasesAttached.DatabaseByName(
  const Name: String): TSqlitePassDatabaseAttached;
begin
  Result := FindDatabase(Name);
  if Result = nil
     then DatabaseErrorFmt(Msg1800, [Name], nil);
end;

destructor TSqlitePassDatabasesAttached.Destroy;
begin

end;

function TSqlitePassDatabasesAttached.FindDatabase(
  const Value: string): TSqlitePassDatabaseAttached;
var
  I: Integer;
begin
  for I := 0 to Pred(Count) do
  begin
    Result := TSqlitePassDatabaseAttached(inherited Items[I]);
    if AnsiCompareText(Result.Name, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TSqlitePassDatabasesAttached.GetItem(
  Index: Integer): TSqlitePassDatabaseAttached;
begin
 Result:=TSqlitePassDatabaseAttached(Inherited Items[Index]);
end;

procedure TSqlitePassDatabasesAttached.SetItem(Index: Integer;
  Database: TSqlitePassDatabaseAttached);
begin
 Put(Index,Database);
end;

{ TSqlitePassDatabaseAttached }

constructor TSqlitePassDatabaseAttached.Create(AOwner: TObject);
begin
  Inherited Create;
  //********************************************************************* Properties
  FParentDatabase := TSqlitePassDatabase(aOwner);
  FDatabase := '';
  FDatasets := TSqlitePassDatasets.create(FParentDatabase);
  //FDatasets.FAttachedDatabase := Self;
  FTableDefs := TSqlitePassTableDefs.Create(FParentDatabase, Self);
  FTableDefs.FAttachedDatabase := Self;
  FQueryDefs := TSqlitePassQueryDefs.Create(FParentDatabase);
  //FQueryDefs.// FAttachedDatabase := Self;
  FViews := TSqlitePassViews.Create(FParentDatabase);
 // FViews.// FAttachedDatabase := Self;
  FIndexDefs := TSqlitePassDatabaseIndexDefs.Create(FParentDatabase);
  //FIndexDefs.// FAttachedDatabase := Self;
end;

destructor TSqlitePassDatabaseAttached.Destroy;
begin
 FDatasets.Free;
 FTableDefs.Free;
 FQueryDefs.Free;
 FViews.Free;
 FIndexDefs.Free;
 Inherited Destroy;
end;

procedure TSqlitePassDatabaseAttached.SetFDatabase(Value: String);
begin
// TODO
end;

end.

