{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Basic sqlite library wrapper for use with fpc compiler
    Author : Luc DAVID Email: luckylazarus@free.fr
    Dowload : http
    portions of code from Mike Cariotoglou Email: mike@deltasingular.gr
    Copyright (C) 2006

  ---------------------------------------------------------------------------

    20.07.2006 - V0.20
      First public release. For testing and debuging...

  --------------------------------------------------------------------------- }

unit SqlitePassEngine;
{$i SqlitePassDbo.inc}
interface
Uses

{$IFDEF FPC}
variants, LCLIntf,
{$ENDIF}
SysUtils, Classes, Db, Forms, Controls, windows,
SqlitePassApi_v3, SqlitePassErrorLang, SqlitePassUtils;

Type
 PPointer= ^Pointer;
 TSqlitePassExecCallBack=function(acolumns: Integer; aColumnValues, aColumnNames:Ppchar):integer of object;
 PSqlitePassExecCallBack=^TSqlitePassExecCallBack;
 TSqlitePassGetPragmaCallback=function(UserData: Pointer; ColumnCount: Integer; ColumnValues, ColumnNames: PPointer): Integer; cdecl;

TSqlitePassEngine = class;
TMemRecord = Class
    public
    RecordIndex: Integer;
    PrimaryKey: Integer;
    Bookmark: Integer;
    BookMarkFlag: TBookmarkFlag;
    FieldsValue: Array of String;
    constructor Create(FieldsCount: Integer);
    Procedure Assign(Const Source: TMemRecord);
    Procedure AssignFieldsValue(const Source: TMemRecord);
  end;

 TSqlitePassTransaction = class
 Private
  FLevel: Integer;
  Engine: TSqlitePassEngine;
//  FInternalTransaction: Boolean;
//  Procedure CommitInternalTransaction;
  Procedure ResetTransactions(Commit: Boolean);
//  Procedure StartInternalTransaction;
 Public
  Constructor Create(Owner: TSqlitePassEngine);
  Procedure Commit;
  Procedure CommitAll;
  Procedure Rollback;
  Procedure RollbackAll;
  Procedure Start;
  Property  Level: Integer Read FLevel Write FLevel;
 end;

 TSqlitePassEngine = Class
  Private
    p,tail:pchar;
    FConnectionHandle: Pointer;
    FStartTime:Integer;
    FStopTime: Integer;
    FElapsedTime : Double;
    FQueryTimeOut    : Integer;
    FTransaction: TSqlitePassTransaction;
  Public
    Constructor Create;
    Destructor Destroy;
    Property QueryTimeout: Integer Read FQueryTimeout Write FQueryTimeout;
    Property Connection: Pointer read FConnectionHandle;
    // Chrono
    Procedure StartChrono;
    Procedure StopChrono;
    Property ElapsedTime: Double read FElapsedTime;
    //------------- Database ----------------
    function GetPragma(pragmaName: String): integer;
    //------------- Database ----------------
    procedure CreateDatabase(FullName: String);
    procedure OpenDatabase(FullName: String);
    procedure CloseDatabase;
    property Database: Pointer Read FConnectionHandle;
    //----------- Transactions -------------
    Property Transaction: TSqlitePassTransaction Read FTransaction Write FTransaction;
    //-------- Tables and Records -----------
//    function CreateTable(Dataset: TSqlitePassDataset): Boolean;
    procedure PrepareStmt(var Stmt: pointer; const sql: String);
    procedure UnprepareStmt(var Stmt: pointer);
    //------------- Sql Exec ----------------
    function _ExecSql(Sql: string; CallBackFunc: TSqlitePassExecCallBack=nil): Integer;
    function Exec(Sql: string; params: Tparams=nil): Integer;
    procedure CheckResult(i: Integer);
   end;



Implementation
uses SqlitePassDbo;

{ ----- Utilities ----- }
function _ExecCallback(cbParam: PSqlitePassExecCallBack; Columns: Integer;
                       ColumnValues, ColumnNames: Ppchar): integer; cdecl;
begin
try
// Result:=cbparam^(columns,columnValues,columnNames);
except
 result:=1;
end;
end;


{ TMemRecord }

procedure TMemRecord.Assign(const Source: TMemRecord);
begin
RecordIndex := Source.RecordIndex;
PrimaryKey  := Source.PrimaryKey;
Bookmark    := Source.Bookmark;
BookMarkFlag:= Source.BookMarkFlag;
AssignFieldsValue(Source);
end;

procedure TMemRecord.AssignFieldsValue(const Source: TMemRecord);
Var
Idx: Integer;
begin
For Idx := 0 to Pred(Length(FieldsValue))
    do FieldsValue[Idx] := Source.FieldsValue[Idx];
end;

constructor TMemRecord.Create(FieldsCount: Integer);
begin
SetLength(FieldsValue,FieldsCount);
end;


{ TSqliteEngine }

constructor TSqlitePassEngine.Create;
begin
FConnectionHandle:= nil;
FQueryTimeOut := 0;
FTransaction := TSqlitePassTransaction.Create(Self);
end;

destructor TSqlitePassEngine.Destroy;
begin
FTransaction.Free;
end;

procedure TSqlitePassEngine.CheckResult(i: Integer);
begin
if i <> SQLITE_OK
   then raise Exception.CreateFmt('%d:%s',[i,SqliteDbv3_Errmsg(FConnectionHandle)])
end;

function TSqlitePassEngine._ExecSql(Sql: string; CallBackFunc: TSqlitePassExecCallBack): Integer;
begin
StartChrono;
repeat
 (*if Assigned(CallBackFunc)
    then result:= SqliteDbv3_Exec(FConnectionHandle,Pchar(Sql),@_ExecCallBack,@@CallBackFunc,nil)
    else*) result:= SqliteDbv3_Exec(FConnectionHandle,Pchar(Sql),nil,nil,nil);
until (result=SQLITE_OK) or not (result in [SQLITE_BUSY,SQLITE_LOCKED])
     or ((FQueryTimeOut>=0) and (Integer(GetTickCount)>=FStartTime));
StopChrono;
CheckResult(Result);
end;

function TSqlitePassEngine.Exec(Sql: string; params: Tparams): Integer;
var
pp,ttail:pchar;
Fst:pointer;
begin
 StartChrono;
 pp:=pchar(sql);
//repeat
  fst:=nil;
  CheckResult(SqliteDbv3_prepare(FConnectionHandle,pp,-1,fst,ttail));
//  if assigned(fst) then
// begin
//    bindParams(params);
    SqliteDbv3_step(fst);
    //step; //execute
(* end
  else
  while (tail^<>#0) and (tail^<=' ') do inc(tail);
  if tail^<>#0 then
   begin
   SqliteDbv3_finalize(fst);
    close; multiple statements, so release cursor
    p:=tail;
   end;
 until tail^=#0;*)
SqliteDbv3_Finalize(fst);
StopChrono;
end;

Procedure TSqlitePassEngine.CreateDatabase(FullName:String);
var
NewDatabase: Pointer;
begin
if Not FileExists(FullName)
   then CheckResult(SqliteDbv3_Open(Pchar(FullName),NewDatabase));
end;

procedure TSqlitePassEngine.CloseDatabase;
begin
Transaction.CommitAll;
CheckResult(SqliteDbv3_Close(FConnectionHandle));
end;

procedure TSqlitePassEngine.OpenDatabase(FullName: String);
begin
CheckSqliteDbv3EngineLoaded;
CheckResult(SqliteDbv3_Open(Pchar(FullName),FConnectionHandle));
end;

(*procedure TSqlitePassEngine.bindParamByName(const fst pointer; const param: tparam);
var idx:integer;
begin
 idx:=SqliteDbv3_bind_parameter_index(fst,pchar(param.Name));
 if idx=0 then raise exception.createFmt('parameter not found : %s',[param.Name]);
 BindParam(param,idx-1);
end;*)


procedure TSqlitePassEngine.PrepareStmt(var Stmt: pointer; const sql: String);
begin
if Assigned(Stmt)
   then SqliteDbv3_Reset(Stmt)
   else CheckResult(SqliteDbv3_prepare(FConnectionHandle,Pchar(Sql),-1,Stmt,tail));
end;


procedure TSqlitePassEngine.UnprepareStmt(var Stmt: pointer);
begin
if Assigned(Stmt)
   then begin
        SqliteDbv3_Finalize(Stmt);
        Stmt := nil;
        end;
end;


{ TSqlitPassTransaction }

procedure TSqlitePassTransaction.Commit;
begin
Engine._ExecSql('Commit;');
Dec(FLevel);
end;

procedure TSqlitePassTransaction.CommitAll;
begin
ResetTransactions(True);
end;

constructor TSqlitePassTransaction.Create(Owner: TSqlitePassEngine);
begin
Engine := TSqlitePassEngine(Owner);
FLevel := 0;
end;

procedure TSqlitePassTransaction.ResetTransactions(Commit: Boolean);
var
i: Integer;
begin
For i := 1 to FLevel do
    Case Commit of
         True  : Self.Commit;
         False : Rollback;
         end;

FLevel := 0;
end;

procedure TSqlitePassTransaction.Rollback;
begin
Engine._ExecSql('Rollback;');
Dec(FLevel);
end;

procedure TSqlitePassTransaction.RollbackAll;
begin
ResetTransactions(False);
end;

(*procedure TSqlitePassTransaction.StartInternalTransaction;
begin
if FLevel = 0
   then begin
        FInternalTransaction := True;
        Engine._ExecSql('Begin;');
        end;
end;

procedure TSqlitePassTransaction.CommitInternalTransaction;
begin
if FInternalTransaction = True
   then begin
        FInternalTransaction := False;
        Engine._ExecSql('Commit;');
        end;
end;*)


procedure TSqlitePassTransaction.Start;
begin
Engine._ExecSql('Begin;');
Inc(FLevel);
end;

(*function TSqlitePassEngine.CreateTable(Dataset: TSqlitePassDataset): Boolean;
begin
Translator.CreateTable(Dataset);
end;*)



function TSqlitePassEngine.GetPragma(pragmaName: String): integer;
var
Recordset : TSqlitePassRecordsetCache;
begin
Recordset := TSqlitePassRecordsetCache.Create(Self);
With Recordset do
     begin
     GetRecords(FConnectionHandle, PragmaName, nil, True, False);
     Result := StrToInt(Items[0].FieldsValue[0]);
     Free;
     end;
end;


procedure TSqlitePassEngine.StartChrono;
begin
 FStartTime:=Integer(GetTickCount);
end;

procedure TSqlitePassEngine.StopChrono;
begin
 FStopTime:=Integer(GetTickCount);
 FElapsedTime := (FStopTime - FStartTime)/1000;
end;

end.

