unit testform;

interface

uses
	Windows, SysUtils, Graphics, Classes, Forms, Controls, ToolWin, ComCtrls,
	StdCtrls, ExtCtrls, XPMan, sqlite3;

type
	TMainForm = class(TForm)
		Panel2: TPanel;
		XPManifest1: TXPManifest;
		CoolBar1: TCoolBar;
		Label1: TLabel;
		Grid: TListView;
    Status: TLabel;
    Panel3: TPanel;
    code: TEdit;
    name: TEdit;
    GoButton: TButton;
    Panel1: TPanel;
    Realtime: TCheckBox;
    LimitRecords: TCheckBox;
    Bevel1: TBevel;
		procedure FormCreate(Sender: TObject);
		procedure FormDestroy(Sender: TObject);
    procedure EditExit(Sender: TObject);
    procedure EditEnter(Sender: TObject);
		procedure GridData(Sender: TObject; Item: TListItem);
    procedure GoButtonClick(Sender: TObject);
    procedure RealtimeClick(Sender: TObject);
    procedure RealtimeUpdate(Sender: TObject);
	private
		db: TSqliteDatabase;
		rs: TStringList;
	end;

var
	MainForm: TMainForm;

implementation

{$R *.dfm}

procedure TMainForm.FormCreate(Sender: TObject);
begin
	db := TSqliteDatabase.Open('unspsc.db3');
	rs := TStringList.Create;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
	rs.Free;
	db.Free;
end;

procedure TMainForm.EditEnter(Sender: TObject);
begin
	Status.Font.Color := clWindowText;
	Status.Caption := 'Type a product ' + TEdit(Sender).Name;
end;

procedure TMainForm.EditExit(Sender: TObject);
begin
	if ActiveControl <> GoButton then
		TEdit(Sender).Text := '';
end;

procedure TMainForm.RealtimeClick(Sender: TObject);
begin
	GoButton.Enabled := not Realtime.Checked;
end;

procedure TMainForm.RealtimeUpdate(Sender: TObject);
begin
	if Realtime.Checked then
		GoButtonClick(Sender);
end;

//	Use OwnerData on the ListView because loading data directly to the grid
//	is slower than the actual query. StringGrid & DrawGrid are also slow.
//
procedure TMainForm.GridData(Sender: TObject; Item: TListItem);
begin
	with Item do begin
		if Index < rs.Count then begin
			Caption := rs.Names[Index];
			if SubItems.Count = 0 then
				SubItems.Add('');
			SubItems[0] := rs.ValueFromIndex[Index];
		end;
	end;
end;

procedure TMainForm.GoButtonClick(Sender: TObject);
var
	sql,field,filter: string;
begin
	if not (Sender is TEdit) then
		if name.Text <> ''
			then Sender := name
			else Sender := code;

	Grid.Clear;
	filter := Trim(TEdit(Sender).Text);
	if filter = '' then begin
		Status.Caption := '';
		Grid.Width := Grid.Width - 1;
		exit;
	end;

	field := TEdit(Sender).Name;
	sql := 'select code,name from unspsc where '
		+ db.Like(field, filter, field = 'code');

	if LimitRecords.Checked then
		sql := sql + ' limit ' + IntToStr(Grid.VisibleRowCount);

	rs.Clear;
	with db.Query(sql) do begin
		while not Eof do begin
			rs.Add(FieldAsString(0) + '=' + FieldAsString(1));
			Next;
		end;
		Free;
	end;

	if rs.Count = 0
		then Status.Font.Color := clRed
		else Status.Font.Color := clWindowText;

	if LimitRecords.Checked and (rs.Count = Grid.VisibleRowCount)
		then Status.Caption := 'Displaying first ' + IntToStr(rs.Count) + ' records'
		else Status.Caption := IntToStr(rs.Count) + ' records found';

	Grid.Items.Count := rs.Count;
	Grid.Width := Grid.Width - 1;
end;

end.
